#include <limits.h>
#include <string.h>
#include <stdio.h>
#include "ssl.h"
#include "s3_state_mon_common_logic.h"
#include "s3_state_mon.h"
//#include <openssl/buffer.h>
//#include <openssl/rand.h>
//#include <openssl/objects.h>
//#include <openssl/evp.h>
//#include <openssl/x509.h>

/*
 * Wrapper for the memcpy function in order to give it some logic.
 */
unsigned char *memcpy_uchar(unsigned char *dest, unsigned char const *src,
		size_t n) {
	return (unsigned char*) memcpy((unsigned char*) dest,
			(unsigned char* const ) src, n);
}


/*@
 assigns \nothing;

 ensures (isClientHello(msg, len, contentType) 								<==> \result == C_Hello) 				&&
 	 	 (isServerHello(msg, len, contentType) 								<==> \result == S_Hello )				&&
 	 	 (isServerCertificate(state, msg, len, direction, contentType) 		<==> \result == S_Cert) 				&&
 	 	 (isServerKeyExchange(msg, len, contentType) 						<==> \result == S_KE) 					&&
 	 	 (isServerCertificateRequest(msg, len, contentType) 				<==> \result == S_CReq) 				&&
 	 	 (isServerHelloDone(msg, len, contentType) 							<==> \result == S_HD )					&&
 	 	 (isClientCertificate(state, msg, len, direction, contentType) 		<==> \result == C_Cert) 				&&
 	 	 (isClientKeyExchange(msg, len, contentType) 						<==> \result == C_KE) 					&&
 	 	 (isClientCertificateVerify(msg, len, contentType) 					<==> \result == C_CV )					&&
 	 	 (isServerChangeCipherSpecs(state, msg, len, direction, contentType)<==> \result ==	S_CCS) 					&&
 	 	 (isServerFin(state, msg, len, direction, contentType) 				<==> \result == S_Fin) 					&&
 	 	 (isClientChangeCipherSpecs(state, msg, len, direction, contentType)<==> \result == C_CCS )					&&
 	 	 (isClientFin(state, msg, len, direction, contentType) 				<==> \result == C_Fin) 					&&
 	 	 (isServerNewSessionTicket(msg, len, contentType) 					<==> \result == ServerNewSessionTicket) &&
 	 	 (isUnknownMessage(state, msg, len, direction, contentType)			<==> \result == UNDEFINED_TYPE )		;
   */
Msg_type getMessageType(STATE_MON *state, unsigned char *msg, unsigned int len,
		int direction, unsigned char contentType);

/*@
 requires isMessageLength(msg, len);

 assigns \nothing;

 ensures
 	 (\result == SSL3 <==> isSSL3(msg, len)) &&
 	 (\result == TLS1 <==> isTLS1_0(msg, len)) &&
 	 (\result == TLS1_1 <==> isTLS1_1(msg, len)) &&
 	 (\result == TLS1_2 <==> isTLS1_2(msg, len)) &&
 	 //(\result == TICKETS <==> isTickets(msg, len)) &&
 	 //(\result == ALPN <==> isALPN(msg, len)) &&
 	 //(\result == DTLS <==> isDTLS(msg, len)) &&
 	 (\result == UNDEFINED_PV <==> isUnknownPV(msg, len));
 */
PV getPV(unsigned char *msg, unsigned int len);

/*@
 requires isMessageLength(msg, len);

 assigns \nothing;

  ensures
 	 (\result == DiffieHellman 	<==> isDH(msg, len))	  		&&
 	 (\result == DHE 			<==> isDHE(msg, len)) 			&&
 	 (\result == DHEExport		<==> isDHEExport(msg, len))		&&
 	 (\result == DHanon 		<==> isDHanon(msg, len))		&&
 	 (\result == ECDH 			<==> isECDH(msg, len))  		&&
 	 (\result == ECDHE 			<==> isECDHE(msg, len)) 		&&
 	 (\result == ECDHanon 		<==> isECDHanon(msg, len)) 		&&
 	 (\result == CS_RSA 		<==> isRSA(msg, len)) 			&&
 	 (\result == RSAExport 		<==> isRSAExport(msg, len))  	 ;
 */
KEM getKEM(unsigned char *msg, unsigned int len);

/*@
  requires isClientCertificate(state, msg, len, direction, handshake);

  assigns \nothing;
  behavior other_certificate:
  	  assumes (msg[1] != 0 || msg[2] != 0 || msg[3] != 0);
  	  ensures \result == AUTH_PROVIDED <==> isNonEmptyClientCertificate{Pre}(state, msg, len, direction);
  behavior empty_certificate :
  	  assumes msg[1] == 0 && msg[2] == 0 && msg[3] == 0;
  	  ensures \result == AUTH_NOT_PROVIDED <==> isEmptyClientCertificate{Pre}(state, msg, len, direction);
  complete behaviors other_certificate, empty_certificate;
  disjoint behaviors other_certificate, empty_certificate;
 */
Auth checkClientCertificate(STATE_MON *state, unsigned char *msg,
		unsigned int len, int direction);

/*
 * Takes the current state, the last message received and its length and appends the message to the log of
 * the state monitor.
 */
/*@
 requires \valid(state) && \valid(msg+(0..len-1)) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(state->log+(0..state->log_length+len-1), state+(0..sizeof(state)-1), msg+(0..len-1));

 ensures MessageAddedToLog{Here,Pre}(state, msg, len);

 assigns state->log[\at(state->log_length,Pre)..\at(state->log_length,Pre)+len-1],
 	 	 state->log_length;
 */
void appendMessageToLog(STATE_MON *state, unsigned char *msg, unsigned int len){
	memcpy_uchar(state->log + state->log_length, msg, len);
	state->log_length += len;
}

/*
 * Handles the ClientHello messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isClientHello(msg, len, handshake);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success :
 	 assumes  StateAfterInitialState(state);

 	 ensures \result == ACCEPT <==> StateAfterClientHello{Here}(state);

 behavior failure :
 	 assumes !(StateAfterInitialState(state));

 	 ensures \result == REJECT;

 disjoint behaviors success, failure;
 complete behaviors success, failure;
  */
int handle_client_hello(STATE_MON *state, unsigned char *msg, unsigned int len, int direction){

	if ( state->version == UNDEFINED_PV 		&&
		 state->role == UNDEFINED_ROLE 			&&
		 state->kem == UNDEFINED_KEM 			&&
		 state->last_message == UNDEFINED_TYPE 	&&
		 state->log_length == 0 				&&
		 state->client_auth == UNDEFINED_AUTH 	&&
		 state->resumption == UNDEFINED 		&&
		 state->renegotiation == UNDEFINED 		&&
		 state->ntick == UNDEFINED) {

		//@ assert StateAfterInitialState{Pre}(state);

		/*@
		 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
		 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));

		 ensures MessageAddedToLog{Here,Pre}(state, msg, len);
		 ensures state->role == CLIENT || state->role == SERVER;
		 ensures state->last_message == C_Hello;

		 assigns state->log_length, state->log[\at(state->log_length,Pre)..\at(state->log_length,Pre)+len-1], state->last_message, state->role;
		 */
		{
			appendMessageToLog(state, msg, len);

			if (direction == SEND)
				state->role = CLIENT;
			else
				state->role = SERVER;

			state->last_message = C_Hello;
			//@ assert \at(state->log_length, Pre) + len == \at(state->log_length, Here);
			//@ assert \forall integer i; 0 <= i < len ==> \at(state->log[i + \at(state->log_length,Pre)],Here) == \at(msg[i],Pre);
			//@ assert \forall integer i; 0 <= i < \at(state->log_length, Pre) ==> \at(state->log[i],Here) == \at(state->log[i],Pre);

			//@ assert MessageAddedToLog{Here,Pre}(state, msg, len);
		}

		//@ assert state->last_message == C_Hello ;
		//@ assert HaveSameStateValuesButRole{Here,Pre}(state);
		//@ assert (state->role == CLIENT || state->role == SERVER );
		//@ assert MessageAddedToLog{Here,Pre}(state,msg, len);
		//@ assert isClientHello{Here}(msg, len, handshake) ;
		//@ assert StateAfterInitialState{Pre}(state);

		//@ assert NextStateAfterClientHello{Here,Pre}(state, msg, len);

		//@ assert StateAfterClientHello{Here}(state);

		return ACCEPT;

	} else {
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : Rejected a non-initial state (with state_mon_initialized == %d) : \nversion : %d\nrole : %d\ncs : %d, \nlast_message : %d\n", state->state_mon_initialized, state->version, state->role, state->kem, state->last_message);
		fprintf(stderr, "log_length : %d\nclient_auth : %d\nresumption : %d, \nrenegociation : %d\n", state->log_length, state->client_auth, state->resumption, state->renegotiation);
#endif
		return REJECT;
	}
}

/*
 * Handles the ServerHello messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isServerHello(msg, len, handshake);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success :

 	 assumes StateAfterClientHello{Pre}(state)
 	 	 	 && sentByServer(state, direction)
 	 	 	 && !isUnknownPV(msg, len) && !isUnknownKEM(msg, len);

 	 ensures \result == ACCEPT ==> StateAfterServerHello{Here}(state);
 	 //assigns
 	 //	 state->version,
 	 //	 state->kem,
 	 //	 state->last_message,
 	 //	 state->log[\at(state->log_length,Pre)..(\at(state->log_length,Pre)+len-1)],
 	 //	 state->log_length;

 behavior failure :
	assumes !(StateAfterClientHello{Pre}(state)
 	 	 	 && sentByServer(state, direction)
 	 	 	 && !isUnknownPV(msg, len) && !isUnknownKEM(msg, len) );

 	ensures \result == REJECT;
 	//assigns \nothing;

 complete behaviors success, failure;
 disjoint behaviors success, failure;
 */
int handle_server_hello(STATE_MON *state, unsigned char *msg, unsigned int len, int direction) {

	if (	state->last_message == C_Hello &&
			((state->role == CLIENT && direction == RECEIVE) || (state->role == SERVER && direction == SEND))) {
		//@ assert StateAfterClientHello{Pre}(state);

		PV version = getPV(msg, len);
		KEM kem = getKEM(msg, len);

		if (version == UNDEFINED_PV|| kem == UNDEFINED_KEM) {
			if (version == UNDEFINED_PV){
#ifdef DEBUG
				fprintf(stderr, "STATE MONITOR : handle_server_hello rejected a message because if its version was unknown.\n");
#endif
			}else{
#ifdef DEBUG
				fprintf(stderr, "STATE MONITOR : handle_server_hello rejected a message because if its cipher suite was unknown.\n");
#endif
			}
#ifdef DEBUG
			int i;
			fprintf(stderr, "STATE_MONITOR : wrong message :\n");
			for (i = 0; i < len; i++) {
				fprintf(stderr, "0x%x ", msg[i]);
				if (i % 10 == 0)
					fprintf(stderr, "\n");
			}
#endif
			return REJECT;
		}
			else {
				//@ assert sentByServer(state, direction);

				/*@
				 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
				 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
				 requires !isUnknownPV(msg, len);
				 requires !isUnknownKEM(msg, len);

				 ensures MessageAddedToLog{Here,Pre}(state, msg, len);
				 ensures AgreeOnVersion{Here}(state, msg, len);
				 ensures AgreeOnKEM{Here}(state, msg, len);
				 ensures state->last_message == S_Hello;

				 assigns state->log_length, state->log[\at(state->log_length,Pre)..\at(state->log_length,Pre)+len-1], state->last_message, state->version, state->kem;
				 */
				{
					appendMessageToLog(state, msg, len);

					state->version = version;
					state->kem = kem;
					state->last_message = S_Hello;
					//@ assert \at(state->log_length, Pre) + len == \at(state->log_length, Here);
					//@ assert \forall integer i; 0 <= i < len ==> \at(state->log[i + \at(state->log_length,Pre)],Here) == \at(msg[i],Pre);
					//@ assert \forall integer i; 0 <= i < \at(state->log_length, Pre) ==> \at(state->log[i],Here) == \at(state->log[i],Pre);

					//@ assert MessageAddedToLog{Here,Pre}(state, msg, len);
				}
				//@ assert \at(state->last_message,Pre) == C_Hello ;
				//@ assert \at(state->last_message,Here) == S_Hello;
				//@ assert HaveSameStateValuesButVersionAndCs{Here,Pre}(state);
				//@ assert AgreeOnVersion{Here}(state, msg, len);
				//@ assert AgreeOnKEM{Here}(state, msg, len);

				//@ assert MessageAddedToLog{Here,Pre}(state,msg, len);
				//@ assert isServerHello{Here}(msg, len, handshake);
				//@ assert StateAfterClientHello{Pre}(state);

				//@ assert NextStateAfterServerHello{Here,Pre}(state, msg, len);

				//@ assert StateAfterServerHello{Here}(state);
			return ACCEPT;
		}
	} else {
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : handle_server_hello rejected a message because the previous state was not appropriate to get a server hello.\n");
#endif
		return REJECT;
	}
}

/*
 * Handles the ServerCertificate messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isServerCertificate(state, msg, len, direction, handshake);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success :
 	 assumes sentByServer(state, direction) 	&&
 	 	 	 StateAfterServerHello{Pre}(state)	&&
 	 	 	 CsRequiresServerCertificate{Pre}(state);

 	 ensures \result == ACCEPT ==> StateAfterServerCertificate(state) ;

 	 //assigns
 	 //	 state->resumption,
 	 //	 state->last_message,
 	 //	 state->log_length,
 	 //	 state->log[\at(state->log_length,Pre)..(\at(state->log_length,Pre)+len-1)];

 behavior failure :
 	 assumes !( sentByServer(state, direction) 	&&
 	 	 	 StateAfterServerHello{Pre}(state)	&&
 	 	 	 CsRequiresServerCertificate{Pre}(state));

 	 ensures \result == REJECT;

 	 //assigns \nothing;
 */
int handle_server_certificate(STATE_MON *state, unsigned char *msg, unsigned int len, int direction) {

	if ( ((state->role == CLIENT && direction == RECEIVE) || (state->role == SERVER && direction == SEND)) &&
			state->last_message == S_Hello) {
		//@ assert StateAfterServerHello{Pre}(state);
		if (	state->kem != DiffieHellman &&
					state->kem != DHE &&
					state->kem != DHEExport &&
					state->kem != ECDH &&
					state->kem != ECDHE &&
					state->kem != CS_RSA &&
					state->kem != RSAExport) {
			//@ assert !CsRequiresServerCertificate{Pre}(state);
#ifdef DEBUG
			fprintf(stderr, "The server certificate was rejected because the cipher suite %d did not require it.\n", state->kem);
#endif
			return REJECT;
		} else {
			//@ assert sentByServer(state, direction);
			//@ assert CsRequiresServerCertificate{Pre}(state);

			/*@
				 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
				 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));


				 ensures MessageAddedToLog{Here,Pre}(state, msg, len);
				 ensures state->last_message == S_Cert;
				 ensures state->resumption == NO;

				 assigns state->log_length, state->log[\at(state->log_length,Pre)..\at(state->log_length,Pre)+len-1], state->last_message, state->resumption;
			 */
			{
				appendMessageToLog(state, msg, len);

				state->last_message = S_Cert;
				state->resumption = NO;
				//@ assert \at(state->log_length, Pre) + len == \at(state->log_length, Here);
				//@ assert \forall integer i; 0 <= i < len ==> \at(state->log[i + \at(state->log_length,Pre)],Here) == \at(msg[i],Pre);
				//@ assert \forall integer i; 0 <= i < \at(state->log_length, Pre) ==> \at(state->log[i],Here) == \at(state->log[i],Pre);

				//@ assert MessageAddedToLog{Here,Pre}(state, msg, len);
			}
			//@ assert \at(state->last_message,Pre) == S_Hello ;
			//@ assert \at(state->last_message,Here) == S_Cert;
			//@ assert \at(state->resumption,Here) == NO;
			//@ assert HaveSameStateValuesButResumption{Here,Pre}(state);

			//@ assert MessageAddedToLog{Here,Pre}(state,msg, len);
			//@ assert isServerCertificate{Here}(state, msg, len, direction, handshake);
			//@ assert StateAfterServerHello{Pre}(state);

			//@ assert NextStateAfterServerCertificate{Here,Pre}(state, msg, len, direction);

			//@ assert StateAfterServerCertificate{Here}(state);

			return ACCEPT;
		}
	} else {
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : handle_server_certificate rejected a message of an inappropriate previous state.\n");
#endif
		return REJECT;
	}
}

/*
 * Handles the ServerKeyExchange messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isServerKeyExchange(msg, len, handshake);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success_from_server_certificate :
 	 assumes
 	 	 sentByServer(state, direction) &&
 	 	 StateAfterServerCertificate{Pre}(state) &&
 	 	 (state->kem == DHE || state->kem && ECDHE || state->kem == DHEExport || state->kem == RSAExport);

 	 ensures \result == ACCEPT ==> StateAfterServerKeyExchange{Here}(state) ;

 behavior success_from_server_hello :
 	 assumes
 	 	 sentByServer(state, direction) &&
 	 	 StateAfterServerHello{Pre}(state) &&
 	 	 (state->kem == DHanon || state->kem == ECDHanon);

 	 ensures \result == ACCEPT ==> StateAfterServerKeyExchange{Here}(state) ;

 behavior failure :
 	 assumes !(sentByServer(state, direction) &&
 	 	 StateAfterServerCertificate{Pre}(state) &&
 	 	 (state->kem == DHE || state->kem && ECDHE || state->kem == DHEExport) ) &&
 	 	 !( sentByServer(state, direction) &&
 	 	 StateAfterServerHello{Pre}(state) &&
 	 	 (state->kem == DHanon || state->kem == ECDHanon) );

 	 ensures \result == REJECT;
 */
int handle_server_key_exchange(STATE_MON *state, unsigned char *msg, unsigned int len, int direction) {

	if ( ((state->role == CLIENT && direction == RECEIVE) || (state->role == SERVER && direction == SEND))
			&& (state->kem != DiffieHellman && state->kem != CS_RSA && state->kem != ECDH)) {

		if(state->last_message != S_Hello && state->last_message != S_Cert){
			//@ assert !StateAfterServerHello{Pre}(state) && !StateAfterServerCertificate{Pre}(state);
#ifdef DEBUG
			fprintf(stderr, "STATE MONITOR : handle_server_key_exchange rejected the SKE because it did not came after the ServerHello or the ServerCertificate.\n");
#endif
			return REJECT;
		}
		else if(state->last_message == S_Hello && state->kem != DHanon && state->kem != ECDHanon){
			//@ assert StateAfterServerHello{Pre}(state) && !(state->kem == DHanon || state->kem == ECDHanon);
#ifdef DEBUG
			fprintf(stderr, "STATE MONITOR : handle_server_key_exchange rejected a message because it came after a server hello and with a wrong CipherSuite.\n");
#endif
			return REJECT;
		}
		else if(state->last_message == S_Cert && state->kem != DHE && state->kem != DHEExport && state->kem != ECDHE && state->kem != RSAExport){
			//@ assert StateAfterServerCertificate{Pre}(state) && !(( state->kem == DHE || state->kem == ECDHE || state->kem == DHEExport || state->kem == RSAExport));
#ifdef DEBUG
			fprintf(stderr, "STATE MONITOR : handle_server_key_exchange rejected a message because it came after a ServerCertificate and with a wrong cipher suite.\n");
#endif
			return REJECT;
		}
		else {
			L:
			//@ assert sentByServer(state, direction);
			/*@
			 assert (StateAfterServerHello{Pre}(state) && (state->kem == DHanon || state->kem == ECDHanon))
			 	 	||
			 	 	(StateAfterServerCertificate{Pre}(state) && ( state->kem == DHE || state->kem == ECDHE || state->kem == DHEExport || state->kem == RSAExport));
			 */

			/*@
			  requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));

			  ensures MessageAddedToLog{Here,Pre}(state, msg, len);
			  ensures state->last_message == S_KE;
			  ensures (\at(state->kem,Pre) == DHanon || \at(state->kem,Pre) == ECDHanon) ==> state->resumption == NO;
			  ensures (\at(state->kem,Pre) == DHE || \at(state->kem,Pre) == ECDHE || \at(state->kem,Pre) == DHEExport || \at(state->kem,Pre) == RSAExport) ==> \at(state->resumption,Here) == \at(state->resumption,Pre);

			  assigns state->log_length, state->log[\at(state->log_length,Pre)..\at(state->log_length,Pre)+len-1], state->last_message, state->resumption;
			 */
			{
				appendMessageToLog(state, msg, len);

				state->last_message = S_KE;

				if (state->kem == DHanon || state->kem == ECDHanon){
					state->resumption = NO;
				}
				//@ assert \at(state->log_length,Here) == \at(state->log_length,Pre) + len ;
				//@ assert (\forall integer i; 0 <= i < \at(state->log_length,Pre) ==> \at(state->log[i],Pre) == \at(state->log[i],Here)) ;
				//@ assert(\forall integer i; 0 <= i < len ==> \at(state->log[i+\at(state->log_length,Pre)],Here) == \at(msg[i],Pre)) ;

				//@ assert MessageAddedToLog{Here,Pre}(state, msg, len);
			}

			//@ assert \at(state->last_message,Here) == S_KE;
			//@ assert HaveSameStateValuesButResumption{Here,Pre}(state);

			//@ assert MessageAddedToLog{Here,Pre}(state,msg, len);
			//@ assert isServerKeyExchange{Here}(msg, len, handshake);
			/*@
			 assert (StateAfterServerHello{Pre}(state) && ( \at(state->kem,Pre) == DHanon || \at(state->kem,Pre) == ECDHanon ) && \at(state->resumption,Here) == NO)
					||
					(StateAfterServerCertificate{Pre}(state) && ( \at(state->kem,Pre) == DHE || \at(state->kem,Pre) == ECDHE || \at(state->kem,Pre) == DHEExport || \at(state->kem,Pre) == RSAExport) && \at(state->resumption,Here) == \at(state->resumption,Pre));
			 */
			//@ assert NextStateAfterServerKeyExchange{Here,Pre}(state, msg, len);
			//@ assert StateAfterServerKeyExchange{Here}(state);

			return ACCEPT;
		}
	} else{
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : handle_server_key_exchange rejected a message because it came at an inappropriate moment.\n");
#endif
		return REJECT;
	}
}

/*
 * Handles the ServerCertificateRequest messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isServerCertificateRequest(msg, len, handshake);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success_after_server_certificate :
 	 assumes StateAfterServerCertificate{Pre}(state)
 	  	  	 && sentByServer(state, direction);

 	 ensures \result == ACCEPT ==> StateAfterServerCertificateRequest{Here}(state) ;

 behavior success_after_server_key_exchange :
 	 assumes StateAfterServerKeyExchange{Pre}(state)
 	 	 	 && (\at(state->kem,Pre) == DiffieHellman ||
				\at(state->kem,Pre) == ECDH ||
				\at(state->kem,Pre) == DHE ||
				\at(state->kem,Pre) == ECDHE ||
				\at(state->kem,Pre) == RSAExport)
 	  	  	 && sentByServer(state, direction);

 	 ensures \result == ACCEPT ==> StateAfterServerCertificateRequest{Here}(state) ;

 behavior failure :
 	 assumes
 	 	 !(
 	 	 StateAfterServerCertificate{Pre}(state)
 	  	  	 && sentByServer(state, direction)
 	  	  	 ) &&
 	 	 !(
 	 	 StateAfterServerKeyExchange{Pre}(state)
 	 	 	 && (\at(state->kem,Pre) == DiffieHellman ||
				\at(state->kem,Pre) == ECDH ||
				\at(state->kem,Pre) == DHE ||
				\at(state->kem,Pre) == ECDHE ||
				\at(state->kem,Pre) == RSAExport)
 	  	  	 && sentByServer(state, direction)
 	 	 );

 	 ensures \result == REJECT;
 */
int handle_server_certificate_request(STATE_MON *state, unsigned char *msg, unsigned int len, int direction){

	if (
			(
			  (state->last_message == S_Cert) ||
			  (state->last_message == S_KE &&
					  (state->kem == DiffieHellman 	||
						state->kem == ECDH 			||
						state->kem == DHE 			||
						state->kem == ECDHE))
			)&&
			((state->role == CLIENT && direction == RECEIVE)
											|| (state->role == SERVER && direction == SEND))) {
			/*@
			 assert StateAfterServerCertificate{Pre}(state) ||
			 ( StateAfterServerKeyExchange{Pre}(state) && (\at(state->kem,Pre) == DiffieHellman ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Pre) == ECDH ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Pre) == DHE ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Pre) == ECDHE) ) ;
			 */

			//@ assert sentByServer(state, direction);


			/*@
			  requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));

			  ensures MessageAddedToLog{Here,Pre}(state, msg, len);
			  ensures state->last_message == S_CReq;
			  ensures state->client_auth == AUTH_REQUESTED;

			  assigns state->log_length, state->log[\at(state->log_length,Pre)..\at(state->log_length,Pre)+len-1], state->last_message, state->client_auth;
			 */
			{
				appendMessageToLog(state, msg, len);

				state->last_message = S_CReq;
				state->client_auth = AUTH_REQUESTED;
				//@ assert \at(state->log_length,Here) == \at(state->log_length,Pre) + len ;
				//@ assert (\forall integer i; 0 <= i < \at(state->log_length,Pre) ==> \at(state->log[i],Pre) == \at(state->log[i],Here)) ;
				//@ assert(\forall integer i; 0 <= i < len ==> \at(state->log[i+\at(state->log_length,Pre)],Here) == \at(msg[i],Pre)) ;

				//@ assert MessageAddedToLog{Here,Pre}(state, msg, len);
			}

			//@ assert \at(state->last_message,Here) == S_CReq;
			//@ assert HaveSameStateValuesButClientAuth{Here,Pre}(state);
			//@ assert \at(state->client_auth,Here) == AUTH_REQUESTED;

			//@ assert MessageAddedToLog{Here,Pre}(state,msg, len);
			//@ assert isServerCertificateRequest{Here}(msg, len, handshake);
			/*@
			 assert ( StateAfterServerCertificate{Pre}(state) )
		  	  ||
		  	  ( StateAfterServerKeyExchange{Pre}(state) && (\at(state->kem,Pre) == DiffieHellman ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Pre) == ECDH ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Pre) == DHE ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Pre) == ECDHE) );
			 */
			//@ assert NextStateAfterServerCertificateRequest{Here,Pre}(state, msg, len);

			//@ assert StateAfterServerCertificateRequest{Here}(state);

			return ACCEPT;

	} else {
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : Rejected because of a wrong state or a wrong direction of the message.\n");
#endif
		return REJECT;
	}
}

/*
 * Handles the ServerHelloDone messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isServerHelloDone(msg, len, handshake);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success_after_server_certificate_request :
 	 assumes StateAfterServerCertificate{Pre}(state)
 	 	 	 && (\at(state->kem,Pre) == DiffieHellman ||
				\at(state->kem,Pre) == ECDH ||
				\at(state->kem,Pre) == CS_RSA)
 	  	  	  && sentByServer(state, direction);

 	 ensures \result == ACCEPT ==> NextStateAfterServerHelloDone{Here,Pre}(state, msg, len) ;

 behavior success_after_server_key_exchange :
	assumes StateAfterServerKeyExchange{Pre}(state)
			 && (\at(state->kem,Pre) == DHanon ||
				\at(state->kem,Pre) == ECDHanon ||
				\at(state->kem,Pre) == DHE ||
				\at(state->kem,Pre) == ECDHE ||
				\at(state->kem,Pre) == DHEExport ||
				\at(state->kem,Pre) == RSAExport)
 	  	  	  && sentByServer(state, direction);

 	ensures \result == ACCEPT ==> NextStateAfterServerHelloDone{Here,Pre}(state, msg, len) ;

 behavior success_after_server_certificate :
	assumes StateAfterServerCertificateRequest{Pre}(state)
			&& (\at(state->kem,Pre) == DiffieHellman ||
			  \at(state->kem,Pre) == ECDH ||
			  \at(state->kem,Pre) == CS_RSA ||
			  \at(state->kem,Pre) == RSAExport ||
			  \at(state->kem,Pre) == DHE ||
			  \at(state->kem,Pre) == DHEExport ||
			  \at(state->kem,Pre) == ECDHE )
 	  	  	  && sentByServer(state, direction);

 	ensures \result == ACCEPT ==> NextStateAfterServerHelloDone{Here,Pre}(state, msg, len) ;

 behavior failure :
 	 assumes
 	 	 !(
			StateAfterServerCertificate{Pre}(state)
 	 	 	 && (\at(state->kem,Pre) == DiffieHellman ||
				\at(state->kem,Pre) == ECDH ||
				\at(state->kem,Pre) == CS_RSA)
 	  	  	  && sentByServer(state, direction)
 	 	 ) && !(
 	 	 	 StateAfterServerKeyExchange{Pre}(state)
			 && (\at(state->kem,Pre) == DHanon ||
				\at(state->kem,Pre) == ECDHanon ||
				\at(state->kem,Pre) == DHE ||
				\at(state->kem,Pre) == ECDHE ||
				\at(state->kem,Pre) == DHEExport ||
				\at(state->kem,Pre) == RSAExport)
 	  	  	  && sentByServer(state, direction)
 	 	 ) && !(
 	 	 	 StateAfterServerCertificateRequest{Pre}(state)
			&& (\at(state->kem,Pre) == DiffieHellman ||
			  \at(state->kem,Pre) == ECDH ||
			  \at(state->kem,Pre) == CS_RSA ||
			  \at(state->kem,Pre) == RSAExport ||
			  \at(state->kem,Pre) == DHE ||
			  \at(state->kem,Pre) == DHEExport ||
			  \at(state->kem,Pre) == ECDHE )
 	  	  	  && sentByServer(state, direction)
 	 	 );

 	 ensures \result == REJECT;
 */
int handle_server_hello_done(STATE_MON *state, unsigned char *msg, unsigned int len, int direction){

	if ( (state->last_message == S_CReq || state->last_message == S_KE || state->last_message == S_Cert) &&
			((state->role == CLIENT && direction == RECEIVE) || (state->role == SERVER && direction == SEND))) {
			//@ assert sentByServer(state,direction);
			//@ assert ( StateAfterServerCertificate{Pre}(state) || StateAfterServerKeyExchange{Pre}(state) || StateAfterServerCertificateRequest{Pre}(state)) ;
		if (state->last_message == S_Cert && state->kem != DiffieHellman && state->kem != ECDH && state->kem != CS_RSA){
#ifdef DEBUG
			fprintf(stderr, "STATE MONITOR : Rejected because after after a server certificate DH, ECDH or RSA is needed to go to server hello done.\n");
#endif
			return REJECT;
		} else if (state->last_message == S_KE && state->kem != DHanon && state->kem != ECDHanon && state->kem != DHE && state->kem != ECDHE && state->kem != DHEExport && state->kem != RSAExport){
#ifdef DEBUG
			fprintf(stderr, "STATE MONITOR : Rejected because after after a server key exchange DHanon, ECDHanon, DHE, DHEExport, RSAExport or ECDHE is needed to go to server hello done.\n");
#endif
			return REJECT;
		} else if (state->last_message == S_CReq && state->kem != DiffieHellman && state->kem != ECDH && state->kem != CS_RSA && state->kem != DHE && state->kem != ECDHE && state->kem != DHEExport && state->kem != RSAExport){
#ifdef DEBUG
			fprintf(stderr, "STATE MONITOR : Rejected because after after a server certificate request DH, ECDH, RSA, DHE, ECDHE or DHEExport is needed to go to server hello done.\n");
#endif
			return REJECT;
		}
		else {
			//@ assert sentByServer(state, direction);
			/*@
			 assert ( StateAfterServerCertificate{Pre}(state) && (\at(state->kem,Pre) == DiffieHellman ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Pre) == ECDH ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Pre) == CS_RSA))
		  	 ||
		  	 ( StateAfterServerKeyExchange{Pre}(state) && (\at(state->kem,Pre) == DHanon ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Pre) == ECDHanon ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Pre) == DHE ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Pre) == DHEExport ||
		  	  	  	  	  	  	  	  	  	  	  	  	    \at(state->kem,Pre) == RSAExport ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Pre) == ECDHE))
		  	 ||
		  	 ( StateAfterServerCertificateRequest{Pre}(state) && (\at(state->kem,Pre) == DiffieHellman ||
																  \at(state->kem,Pre) == ECDH ||
																  \at(state->kem,Pre) == CS_RSA ||
																  \at(state->kem,Pre) == DHE ||
																  \at(state->kem,Pre) == DHEExport ||
																  \at(state->kem,Pre) == RSAExport ||
																  \at(state->kem,Pre) == ECDHE ));
			 */


			/*@
				requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));

			  ensures MessageAddedToLog{Here,Pre}(state, msg, len);
			  ensures state->last_message == S_HD;
			  ensures (\at(state->last_message,Pre) == S_Cert || \at(state->last_message,Pre) == S_KE) ==> state->client_auth == NO_AUTH;
			  ensures (\at(state->last_message,Pre) == S_CReq) ==> ( state->client_auth == \at(state->client_auth,Pre) );
			  assigns state->log_length, state->log[\at(state->log_length,Pre)..\at(state->log_length,Pre)+len-1], state->last_message, state->client_auth;
			 */
			{
				appendMessageToLog(state, msg, len);

				if(state->last_message == S_Cert || state->last_message == S_KE)
					state->client_auth = NO_AUTH;

				state->last_message = S_HD;
				//@ assert \at(state->log_length,Here) == \at(state->log_length,Pre) + len ;
				//@ assert (\forall integer i; 0 <= i < \at(state->log_length,Pre) ==> \at(state->log[i],Pre) == \at(state->log[i],Here)) ;
				//@ assert(\forall integer i; 0 <= i < len ==> \at(state->log[i+\at(state->log_length,Pre)],Here) == \at(msg[i],Pre)) ;

				//@ assert MessageAddedToLog{Here,Pre}(state, msg, len);

			}
			//@ assert \at(state->last_message,Here) == S_HD;
			//@ assert HaveSameStateValuesButClientAuth{Here,Pre}(state);

			//@ assert MessageAddedToLog{Here,Pre}(state,msg, len);
			//@ assert isServerHelloDone{Here}(msg, len, handshake);

			//@ assert StateAfterServerCertificate{Pre}(state) ==> \at(state->client_auth,Here) == NO_AUTH;
			//@ assert StateAfterServerKeyExchange{Pre}(state) ==> \at(state->client_auth,Here) == NO_AUTH;
			//@ assert StateAfterServerCertificateRequest{Pre}(state) ==> \at(state->client_auth,Here) == \at(state->client_auth,Pre);

			/*@
			 assert ( StateAfterServerCertificate{Pre}(state) && (\at(state->kem,Pre) == DiffieHellman ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Pre) == ECDH ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Pre) == CS_RSA)
		  	  	  	  	  	  	  	  	  	  	  	   && \at(state->client_auth,Here) == NO_AUTH)
		  	 ||
		  	 ( StateAfterServerKeyExchange{Pre}(state) && (\at(state->kem,Pre) == DHanon ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Pre) == ECDHanon ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Pre) == DHE ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Pre) == DHEExport ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Pre) == RSAExport ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Pre) == ECDHE)
		  	  	  	  	  	  	  	  	  	  	  	  	  	&& \at(state->client_auth,Here) == NO_AUTH)
		  	 ||
		  	 ( StateAfterServerCertificateRequest{Pre}(state) && (\at(state->kem,Pre) == DiffieHellman ||
																  \at(state->kem,Pre) == ECDH ||
																  \at(state->kem,Pre) == CS_RSA ||
																  \at(state->kem,Pre) == DHE ||
																  \at(state->kem,Pre) == DHEExport ||
																  \at(state->kem,Pre) == RSAExport ||
																  \at(state->kem,Pre) == ECDHE )
																  && \at(state->client_auth,Here) == \at(state->client_auth,Pre));
			 */
			//@ assert NextStateAfterServerHelloDone{Here,Pre}(state, msg, len);

			//@ assert StateAfterServerHelloDone{Here}(state);

			return ACCEPT;
		}
	} else {
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : Rejected because of a mismatch with the previous state.\n");
#endif
		return REJECT;
	}
}

/*
 * Handles the ClientCertificates messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isClientCertificate(state, msg, len, direction, handshake);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success :
 	 assumes  StateAfterServerHelloDone{Pre}(state)
 	 	 	  && state->client_auth == AUTH_REQUESTED
 	  	  	  && sentByClient(state, direction);

 	 ensures \result == ACCEPT ==> NextStateAfterClientCertificate{Here,Pre}(state, msg, len, direction) ;

 behavior failure :
 	 assumes
 	 	 !(	StateAfterServerHelloDone{Pre}(state)
 	 	 	  && state->client_auth == AUTH_REQUESTED
 	  	  	  && sentByClient(state, direction) );

 	 ensures \result == REJECT;
 */
int handle_client_certificate(STATE_MON *state, unsigned char *msg, unsigned int len, int direction){

	if (	state->last_message == S_HD
			&& state->client_auth == AUTH_REQUESTED
			&& ((state->role == CLIENT && direction == SEND)
								|| (state->role == SERVER && direction == RECEIVE))) {

		//@ assert sentByClient(state, direction);
		//@ assert StateAfterServerHelloDone{Pre}(state);
		//@ assert \at(state->client_auth,Pre) == AUTH_REQUESTED;

		/*@
		  requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
		  requires isClientCertificate(state, msg, len, direction, handshake);

		  ensures MessageAddedToLog{Here,Pre}(state, msg, len);
		  ensures state->last_message == C_Cert;
		  ensures (isEmptyClientCertificate{Pre}(state, msg, len, direction) && \at(state->client_auth,Here) == AUTH_NOT_PROVIDED )
					||
					(isNonEmptyClientCertificate{Pre}(state, msg, len, direction) && \at(state->client_auth,Here) == AUTH_PROVIDED );

		  assigns state->log_length, state->log[\at(state->log_length,Pre)..\at(state->log_length,Pre)+len-1], state->client_auth, state->last_message;
		 */
		{
			appendMessageToLog(state, msg, len);
			state->client_auth = checkClientCertificate(state, msg, len, direction);

			state->last_message = C_Cert;

			//@ assert \at(state->log_length,Pre) + len == \at(state->log_length,Here);
			//@ assert \forall integer i; 0 <= i < \at(state->log_length, Pre) ==> \at(state->log[i],Pre) == \at(state->log[i],Here);
			//@ assert \forall integer i; 0 <= i < len ==> \at(state->log[i + \at(state->log_length,Pre)],Here) == \at(msg[i],Pre);

			//@ assert MessageAddedToLog{Here,Pre}(state, msg, len);
		}
		//@ assert \at(state->last_message,Here) == C_Cert;
		//@ assert HaveSameStateValuesButClientAuth{Here,Pre}(state);

		//@ assert MessageAddedToLog{Here,Pre}(state,msg, len);
		//@ assert isClientCertificate{Here}(state, msg, len, direction, handshake);
		//@ assert StateAfterServerHelloDone{Pre}(state);
		//@ assert \at(state->client_auth,Pre) == AUTH_REQUESTED;

		//@ assert NextStateAfterClientCertificate{Here,Pre}(state, msg, len, direction);
		//@ assert StateAfterClientCertificate{Here}(state);
		return ACCEPT;

	} else {
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : Rejected because of a mismatch with the previous state.\n");
#endif
		return REJECT;
	}
}

/*
 * Handles the ClientKeyExchange messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isClientKeyExchange(msg, len, handshake);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success_after_server_hello_done :
 	 assumes  StateAfterServerHelloDone{Pre}(state)
 	 	 	  && state->client_auth == NO_AUTH
 	  	  	  && sentByClient(state, direction);

 	 ensures \result == ACCEPT ==> NextStateAfterClientKeyExchange{Here,Pre}(state, msg, len) ;

 behavior success_after_client_certificate :
 	 assumes StateAfterClientCertificate{Pre}(state) &&
			 (state->client_auth == AUTH_PROVIDED || state->client_auth == AUTH_NOT_PROVIDED)
 	  	  	 && sentByClient(state, direction);

	ensures \result == ACCEPT ==> NextStateAfterClientKeyExchange{Here,Pre}(state, msg, len) ;

 behavior failure :
 	 assumes
 	 	 !(
		 StateAfterServerHelloDone{Pre}(state)
 	 	 	  && state->client_auth == NO_AUTH
 	  	  	  && sentByClient(state, direction)
 	 	 ) &&
 	 	 !(
 	 	 StateAfterClientCertificate{Pre}(state) &&
			 (state->client_auth == AUTH_PROVIDED || state->client_auth == AUTH_NOT_PROVIDED)
 	  	  	 && sentByClient(state, direction)
 	 	 );

 	 ensures \result == REJECT;
 */
int handle_client_key_exchange(STATE_MON *state, unsigned char *msg, unsigned int len, int direction){

	if (
			((state->last_message == S_HD
				&& state->client_auth == NO_AUTH
			  ) ||
			  (state->last_message == C_Cert &&
				 ( state->client_auth == AUTH_PROVIDED || state->client_auth == AUTH_NOT_PROVIDED ))
			) &&
			((state->role == CLIENT && direction == SEND)
											|| (state->role == SERVER && direction == RECEIVE))) {

			/*@
			 assert (StateAfterClientCertificate{Pre}(state) && (\at(state->client_auth,Pre) == AUTH_PROVIDED || \at(state->client_auth,Pre) == AUTH_NOT_PROVIDED))
					||
					(StateAfterServerHelloDone{Pre}(state) && ( \at(state->client_auth,Pre) == NO_AUTH ));
			 */
			//@ assert sentByClient(state, direction);

			/*@
			  requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));

			  ensures MessageAddedToLog{Here,Pre}(state, msg, len);
			  ensures state->last_message == C_KE;
			  assigns state->log_length, state->log[\at(state->log_length,Pre)..\at(state->log_length,Pre)+len-1], state->last_message;
			 */
			{
				appendMessageToLog(state, msg, len);
				state->last_message = C_KE;
				//@ assert \at(state->log_length,Here) == \at(state->log_length,Pre) + len ;
				//@ assert (\forall integer i; 0 <= i < \at(state->log_length,Pre) ==> \at(state->log[i],Pre) == \at(state->log[i],Here)) ;
				//@ assert(\forall integer i; 0 <= i < len ==> \at(state->log[i+\at(state->log_length,Pre)],Here) == \at(msg[i],Here)) ;

				//@ assert MessageAddedToLog{Here,Pre}(state, msg, len);
			}
			//@ assert \at(state->last_message,Here) == C_KE;
			//@ assert HaveSameStateValuesAndRole{Here,Pre}(state);


			//@ assert MessageAddedToLog{Here,Pre}(state,msg, len);
			//@ assert isClientKeyExchange{Here}(msg, len, handshake);
			/*@
			 assert (StateAfterClientCertificate{Pre}(state) && (\at(state->client_auth,Pre) == AUTH_PROVIDED || \at(state->client_auth,Pre) == AUTH_NOT_PROVIDED))
					||
					(StateAfterServerHelloDone{Pre}(state) && ( \at(state->client_auth,Pre) == NO_AUTH ));
			 */

			//@ assert NextStateAfterClientKeyExchange{Here,Pre}(state, msg, len);

			//@ assert StateAfterClientKeyExchange{Here}(state);

			return ACCEPT;

	} else {
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : Rejected because the state does not match what expected.\n");
#endif
		return REJECT;
	}
}

/*
 * Handles the ClientCertificateVerify messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isClientCertificateVerify(msg, len, handshake);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success :
 	 assumes  StateAfterClientKeyExchange{Pre}(state)
 	  	  	  && sentByClient(state, direction)
 	  	  	  && \at(state->client_auth,Pre) == AUTH_PROVIDED;

 	 ensures \result == ACCEPT ==> StateAfterClientCertificateVerify{Here}(state);

 behavior failure :
 	 assumes !( StateAfterClientKeyExchange{Pre}(state)
 	  	  	  && sentByClient(state, direction)
 	  	  	  && \at(state->client_auth,Pre) == AUTH_PROVIDED);

 	 ensures \result == REJECT;
 */
int handle_client_certificate_verify(STATE_MON *state, unsigned char *msg, unsigned int len, int direction){

	if (   state->last_message == C_KE
			&& state->client_auth == AUTH_PROVIDED
			&& ((state->role == CLIENT && direction == SEND)
								|| (state->role == SERVER && direction == RECEIVE))) {
			//@ assert sentByClient(state, direction);
			//@ assert StateAfterClientKeyExchange{Pre}(state);
			//@ assert \at(state->client_auth,Pre) == AUTH_PROVIDED;

			/*@
			  requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));

			  ensures MessageAddedToLog{Here,Pre}(state, msg, len);
			  ensures state->last_message == C_CV;
			  assigns state->log_length, state->log[\at(state->log_length,Pre)..\at(state->log_length,Pre)+len-1], state->last_message;
			 */
			{
				appendMessageToLog(state, msg, len);

				state->last_message = C_CV;

				//@ assert \at(state->log_length,Pre) + len == \at(state->log_length,Here);
				//@ assert \forall integer i; 0 <= i < \at(state->log_length, Pre) ==> \at(state->log[i],Pre) == \at(state->log[i],Here);
				//@ assert \forall integer i; 0 <= i < len ==> \at(state->log[i + \at(state->log_length,Pre)],Here) == \at(msg[i],Pre);

				//@ assert MessageAddedToLog{Here,Pre}(state, msg, len);
			}
			//@ assert \at(state->last_message,Here) == C_CV;
			//@ assert HaveSameStateValuesAndRole{Here,Pre}(state);

			// @ assert \at(state->log_length,Here) == \at(state->log_length,Pre) + len ;
			// @ assert (\forall integer i; 0 <= i < \at(state->log_length,Pre) ==> \at(state->log[i],Pre) == \at(state->log[i],Here)) ;
			// @ assert(\forall integer i; 0 <= i < len ==> \at(state->log[i+\at(state->log_length,Pre)],Here) == \at(msg[i],Here)) ;

			//@ assert MessageAddedToLog{Here,Pre}(state,msg, len);
			//@ assert isClientCertificateVerify{Here}(msg, len, handshake);
			//@ assert StateAfterClientKeyExchange{Pre}(state);
			//@ assert \at(state->client_auth,Pre) == AUTH_PROVIDED;

			//@ assert NextStateAfterClientCertificateVerify{Here,Pre}(state, msg, len);
			//@ assert StateAfterClientCertificateVerify{Here}(state);

			return ACCEPT;
	} else {
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : Rejected because the state does not match what expected.\n");
#endif
		return REJECT;

	}
}

/*
 * Handles the ServerNewSessionTicket messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isServerNewSessionTicket(msg, len, handshake);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success_after_server_hello :
 	 assumes  StateAfterServerHello{Pre}(state)
 	  	  	  && sentByServer(state, direction);

 	 ensures \result == ACCEPT ==> StateAfterServerNewSessionTicket{Here}(state);

 behavior success_after_client_fin :
 	 assumes  StateAfterClientFin{Pre}(state)
 	  	  	  && sentByServer(state, direction);

 	 ensures \result == ACCEPT ==> StateAfterServerNewSessionTicket{Here}(state);

 behavior failure :
 	 assumes !( StateAfterServerHello{Pre}(state)
 	  	  	  && sentByServer(state, direction)
 	  	  	  ) &&
 	  	  	  !(
 	  	  	  StateAfterClientFin{Pre}(state)
 	  	  	  && sentByServer(state, direction)
 	  	  	  );

 	 ensures \result == REJECT;
 */
int handle_server_new_session_ticket(STATE_MON *state, unsigned char *msg, unsigned int len, int direction){

	if (   ( state->last_message == S_Hello || (state->last_message == C_Fin && state->resumption == NO) )
			&& ((state->role == SERVER && direction == SEND) || (state->role == CLIENT && direction == RECEIVE))) {
		//@ assert sentByServer(state, direction);
		//@ assert ( StateAfterServerHello{Pre}(state)  ) || (StateAfterClientFin{Pre}(state) && \at(state->resumption,Pre) == NO);

		/*@
		  requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));

		  ensures MessageAddedToLog{Here,Pre}(state, msg, len);
		  ensures state->last_message == ServerNewSessionTicket;
		  ensures \at(state->last_message,Pre) == S_Hello ==> (state->resumption == YES);
		  ensures \at(state->last_message,Pre) == C_Fin ==> (state->resumption == NO);
		  ensures state->ntick == YES;
		  assigns state->log_length, state->log[\at(state->log_length,Pre)..\at(state->log_length,Pre)+len-1], state->last_message, state->resumption, state->ntick;
		 */
		{
			appendMessageToLog(state, msg, len);
			if (state->last_message == S_Hello)
				state->resumption = YES;

			state->last_message = ServerNewSessionTicket;
			state->ntick = YES;
			//@ assert \at(state->log_length,Pre) + len == \at(state->log_length,Here);
			//@ assert \forall integer i; 0 <= i < \at(state->log_length, Pre) ==> \at(state->log[i],Pre) == \at(state->log[i],Here);
			//@ assert \forall integer i; 0 <= i < len ==> \at(state->log[i + \at(state->log_length,Pre)],Here) == \at(msg[i],Pre);

			//@ assert MessageAddedToLog{Here,Pre}(state, msg, len);
		}
		//@ assert \at(state->last_message,Here) == ServerNewSessionTicket;
		//@ assert HaveSameStateValuesButResumptionAndTicket{Here,Pre}(state);

		//@ assert MessageAddedToLog{Here,Pre}(state,msg, len);
		//@ assert isServerNewSessionTicket{Here}(msg, len, handshake);
		//@ assert \at(state->ntick,Here) == YES;
		//@ assert ( StateAfterServerHello{Pre}(state) && \at(state->resumption,Here) == YES ) || (StateAfterClientFin{Pre}(state) && \at(state->resumption,Pre) == \at(state->resumption,Here));

		//@ assert NextStateAfterServerNewSessionTicket{Here,Pre}(state, msg, len);
		//@ assert StateAfterServerNewSessionTicket{Here}(state);
		return ACCEPT;
	} else {
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : Rejected because the state does not match what expected.\n");
#endif
		return REJECT;
	}
}

/*
 * Handles the ServerCCS messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isServerChangeCipherSpecs(state, msg, len, direction, change_cipher_spec);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success_after_resumption_no_session_ticket :
 	 assumes StateAfterServerHello{Pre}(state)
 	  	  	  && sentByServer(state, direction);

 	 ensures \result == ACCEPT ==> StateAfterServerCCS{Here}(state);

 behavior success_after_resumption_and_session_ticket :
 	 assumes StateAfterServerNewSessionTicket{Pre}(state)
 	  	  	  && sentByServer(state, direction);

 	 ensures \result == ACCEPT ==> StateAfterServerCCS{Here}(state);

 behavior success_no_resumption :
 	 assumes StateAfterClientFin{Pre}(state)
 	  	 	 && \at(state->resumption,Pre) == NO
 	  	  	 && sentByServer(state, direction);

 	 ensures \result == ACCEPT ==> StateAfterServerCCS{Here}(state) ;

 behavior failure :
 	 assumes
 	 	 !(
 	 	 StateAfterServerHello{Pre}(state)
 	  	  	  && sentByServer(state, direction)
		  ) &&
 	 	 !(
 	 	 StateAfterServerNewSessionTicket{Pre}(state)
 	  	  	  && sentByServer(state, direction)
 	 	 )&&
 	 	 !(
 	 	 StateAfterClientFin{Pre}(state)
 	 	 	 && \at(state->resumption,Pre) == NO
 	  	  	 && sentByServer(state, direction)
 	 	 );

 	 ensures \result == REJECT;
 */
int handle_server_change_cipher_specs(STATE_MON *state, unsigned char *msg, unsigned int len, int direction){

	if (
	    ( state->last_message == S_Hello || (state->last_message == C_Fin && state->resumption == NO) || (state->last_message == ServerNewSessionTicket ) )
			&&
			((state->role == CLIENT && direction == RECEIVE)
											|| (state->role == SERVER && direction == SEND))) {
			// @ assert messageFitsInLog(state, len);
			// @ assert sentByServer(state, direction);
			/*@
			 assert (StateAfterServerHello{Pre}(state) )
					||
					(StateAfterClientFin{Pre}(state) && \at(state->resumption,Pre) == NO )
					||
					(StateAfterServerNewSessionTicket{Pre}(state) );
			 */

		/*@
		 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));

		  ensures state->last_message == S_CCS;
		  ensures (\at(state->last_message,Pre) == S_Hello) ==> state->resumption == YES && state->ntick == NO;
		  ensures (\at(state->last_message,Pre) == C_Fin) ==> state->ntick == NO && state->resumption == NO;
		  ensures (\at(state->last_message,Pre) == ServerNewSessionTicket) ==> (\at(state->resumption,Pre) == state->resumption && \at(state->ntick,Pre) == state->ntick);

		  assigns state->last_message, state->resumption, state->ntick;
		 */
		{

			if(state->last_message == S_Hello){
				state->resumption = YES;
				state->ntick = NO;
			}
			else if (state->last_message == C_Fin){
				state->ntick = NO;
			}

			state->last_message = S_CCS;

		}

			//@ assert \at(state->last_message,Here) == S_CCS;
			//@ assert HaveSameStateValuesButResumptionAndTicket{Here,Pre}(state);

			//@ assert isServerChangeCipherSpecs{Here}(state, msg, len, direction, change_cipher_spec);
		        //@ assert (\at(state->last_message,Pre) == C_Fin && \at(state->resumption,Pre) == NO) ==> ( StateAfterClientFin{Pre}(state) &&  state->ntick == NO && state->resumption == NO);
		        //@ assert (\at(state->last_message,Pre) == S_Hello) ==> ( StateAfterServerHello{Pre}(state) && \at(state->resumption,Here) == YES && \at(state->ntick,Here) == NO);
                        //@ assert (\at(state->last_message,Pre) == ServerNewSessionTicket ) ==> (StateAfterServerNewSessionTicket{Pre}(state) && \at(state->resumption,Here) == \at(state->resumption,Pre) && \at(state->ntick,Here) == \at(state->ntick,Pre));
			/*@
			 assert (StateAfterServerHello{Pre}(state) && \at(state->resumption,Here) == YES && \at(state->ntick,Here) == NO )
					||
					(StateAfterClientFin{Pre}(state) && \at(state->resumption,Pre) == NO && \at(state->resumption,Pre) == \at(state->resumption,Here)  && \at(state->ntick,Here) == NO )
					||
					(StateAfterServerNewSessionTicket{Pre}(state) && \at(state->resumption,Here) == \at(state->resumption,Pre) && \at(state->ntick,Here) == \at(state->ntick,Pre));
			 */

			//@ assert NextStateAfterServerCCS{Here,Pre}(state, msg, len, direction);
			//@ assert StateAfterServerCCS{Here}(state);

			return ACCEPT;
	} else {
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : Rejected because the state does not match what expected.\n");
#endif
		return REJECT;
	}
}

/*
 * Handles the ServerFinished messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isServerFin(state, msg, len, direction, handshake);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success :
 	 assumes StateAfterServerCCS{Pre}(state)
 	  	  	  && sentByServer(state, direction);

 	 ensures \result == ACCEPT ==> StateAfterServerFin{Here}(state) ;

 behavior failure :
 	 assumes !( StateAfterServerCCS{Pre}(state)
 	  	  	  && sentByServer(state, direction) );

 	 ensures \result == REJECT;
 */
int handle_server_fin(STATE_MON *state, unsigned char *msg, unsigned int len, int direction){

	if (  state->last_message == S_CCS &&
		((state->role == CLIENT && direction == RECEIVE) || (state->role == SERVER && direction == SEND))) {
			// @ assert sentByServer(state, direction);
			// @ assert StateAfterServerCCS{Pre}(state) ;

		/*@
		 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));

		  ensures MessageAddedToLog{Here,Pre}(state, msg, len);
		  ensures state->last_message == S_Fin;

		  assigns state->log_length, state->log[\at(state->log_length,Pre)..\at(state->log_length,Pre)+len-1], state->last_message;
		 */
		{
			appendMessageToLog(state, msg, len);
			state->last_message = S_Fin;

			//@ assert \at(state->log_length,Pre) + len == \at(state->log_length,Here);
			//@ assert \forall integer i; 0 <= i < \at(state->log_length, Pre) ==> \at(state->log[i],Pre) == \at(state->log[i],Here);
			//@ assert \forall integer i; 0 <= i < len ==> \at(state->log[i + \at(state->log_length,Pre)],Here) == \at(msg[i],Pre);

			//@ assert MessageAddedToLog{Here,Pre}(state, msg, len);
		}

		//@ assert \at(state->last_message,Here) == S_Fin;
		//@ assert HaveSameStateValuesAndRole{Here,Pre}(state);

		//@ assert MessageAddedToLog{Here,Pre}(state,msg, len);
		//@ assert isServerFin{Here}(state, msg, len, direction, handshake);
		//@ assert StateAfterServerCCS{Pre}(state);

		//@ assert NextStateAfterServerFin{Here,Pre}(state, msg, len, direction);
		//@ assert StateAfterServerFin{Here}(state);

		return ACCEPT;
	} else {
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : Rejected because the state does not match what expected.\n");
#endif
		return REJECT;
	}
}

/*
 * Handles the ClientCCS messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isClientChangeCipherSpecs(state, msg, len, direction, change_cipher_spec);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success_after_resumption :
 	 assumes StateAfterServerFin(state)
 	 	 	  && state->resumption == YES
 	  	  	  && sentByClient(state, direction);

 	 ensures \result == ACCEPT ==> StateAfterClientCCSLastMsg{Here}(state) ;

 behavior success_after_client_certificate_verify :
 	 assumes StateAfterClientCertificateVerify{Pre}(state)
 	  	  	  && sentByClient(state, direction);

 	 ensures \result == ACCEPT ==> StateAfterClientCCS{Here}(state) ;

 behavior success_after_client_key_exchange :
 	 assumes StateAfterClientKeyExchange{Pre}(state)
 	 	 	 && (\at(state->client_auth,Pre) == NO_AUTH || \at(state->client_auth,Pre) == AUTH_NOT_PROVIDED )
 	  	  	 && sentByClient(state, direction);

 	 ensures \result == ACCEPT ==> StateAfterClientCCS{Here}(state) ;

 behavior success_after_client_certificate :
 	 assumes StateAfterClientCertificate{Pre}(state)
 	 	 	 && (\at(state->kem,Pre) == DiffieHellman || \at(state->kem,Pre) == ECDH )
 	  	  	 && sentByClient(state, direction);

 	 ensures \result == ACCEPT ==> StateAfterClientCCS{Here}(state) ;

 behavior failure :
 	 assumes
 	 	 !sentByClient(state, direction)	||
 	 	 (
 	 	 !(
 	 	 StateAfterClientCertificate{Pre}(state)
 	 	 	 && (\at(state->kem,Pre) == DiffieHellman || \at(state->kem,Pre) == ECDH )
		  )
		  	  &&
 	 	 !(
 	 	 StateAfterClientKeyExchange{Pre}(state)
 	 	 	 && (\at(state->client_auth,Pre) == NO_AUTH || \at(state->client_auth,Pre) == AUTH_NOT_PROVIDED )
 	 	 )
 	 	 	 &&
 	 	 !(
 	 	 StateAfterClientCertificateVerify{Pre}(state)
 	 	 )
 	 	 	 &&
 	 	 !(
 	 	 	 StateAfterServerFin(state)
 	 	 	  && state->resumption == YES
 	 	 )
 	 	 );

 	 ensures \result == REJECT;
 */
int handle_client_change_cipher_specs(STATE_MON *state, unsigned char *msg, unsigned int len, int direction){

	if (
			(
					state->last_message == C_CV ||
					(state->last_message == C_KE && (state->client_auth == AUTH_NOT_PROVIDED || state->client_auth == NO_AUTH) ) ||
					(state->last_message == C_Cert && (state->kem == DiffieHellman || state->kem == ECDH) ) ) &&
			((state->role == SERVER && direction == RECEIVE) || (state->role == CLIENT && direction == SEND))) {
		//@ assert sentByClient(state, direction);

		/*@
		 assert (StateAfterClientCertificateVerify{Pre}(state)  )
				||
				(StateAfterClientKeyExchange{Pre}(state) && (\at(state->client_auth,Pre) == NO_AUTH || \at(state->client_auth,Pre) == AUTH_NOT_PROVIDED ))
				||
				(StateAfterClientCertificate{Pre}(state) && (\at(state->kem,Pre) == DiffieHellman || \at(state->kem,Pre) == ECDH) );
		 */

		/*@
		 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));

		  ensures state->last_message == C_CCS;
		  ensures state->resumption == NO;
		  assigns  state->resumption, state->last_message;
		 */
		{

			state->last_message = C_CCS;
			state->resumption = NO;

		}
		//@ assert \at(state->last_message,Here) == C_CCS;
		//@ assert HaveSameStateValuesButResumption{Here,Pre}(state);
		//@ assert \at(state->resumption,Here) == NO;
		//@ assert isClientChangeCipherSpecs{Here}(state, msg, len, direction, change_cipher_spec);
		/*@
		 assert (StateAfterClientCertificateVerify{Pre}(state)  )
				||
				(StateAfterClientKeyExchange{Pre}(state) && (\at(state->client_auth,Pre) == NO_AUTH || \at(state->client_auth,Pre) == AUTH_NOT_PROVIDED ))
				||
				(StateAfterClientCertificate{Pre}(state) && (\at(state->kem,Pre) == DiffieHellman || \at(state->kem,Pre) == ECDH) );
		 */

		//@ assert NextStateAfterClientCCS{Here,Pre}(state, msg, len, direction);
		//@ assert StateAfterClientCCS{Here}(state);

		return ACCEPT;
	}
		else if (
					state->last_message == S_Fin && state->resumption == YES &&
					((state->role == SERVER && direction == RECEIVE) || (state->role == CLIENT && direction == SEND))
				) {
			//@ assert sentByClient(state, direction);
			//@ assert StateAfterServerFin(state) && state->resumption == YES;

			/*@
			 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));

			  ensures state->last_message == C_CCS;
			  assigns state->last_message;
			 */
			{
				state->last_message = C_CCS;

			}

			//@ assert \at(state->last_message,Here) == C_CCS								;
			//@ assert HaveSameStateValuesAndRole{Here,Pre}(state)									;
			//@ assert isClientChangeCipherSpecs{Here}(state, msg, len, direction, change_cipher_spec);
			//@ assert (StateAfterServerFin{Pre}(state) && \at(state->resumption,Here) == YES );
			//@ assert \at(state->resumption,Pre) == YES;
			//@ assert NextStateAfterClientCCSLastMsg{Here,Pre}(state,msg, len, direction);
			//@ assert StateAfterClientCCSLastMsg{Here}(state);
			return ACCEPT;
	} else {
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : Rejected because the state does not match what expected.\n");
#endif
		return REJECT;
	}
}

/*
 * Handles the ClientFinished messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isClientFin(state, msg, len, direction, handshake);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success_last_message :
 	 assumes StateAfterClientCCSLastMsg{Pre}(state)
 	 	 	 && \at(state->resumption,Pre) == YES
 	  	  	  && sentByClient(state, direction);

 	 ensures \result == ACCEPT ==> StateAfterClientFinLastMsg{Here}(state) ;

  behavior success_not_last_message :
 	 assumes StateAfterClientCCS{Pre}(state)
 	 	 	 && \at(state->resumption,Pre) == NO
 	  	  	  && sentByClient(state, direction);

 	 ensures \result == ACCEPT ==> StateAfterClientFin{Here}(state) ;

 behavior failure :
 	 assumes !( ( StateAfterClientCCS{Pre}(state) || StateAfterClientCCSLastMsg{Pre}(state) )
 	  	  	  && sentByClient(state, direction) );

 	 ensures \result == REJECT;
 */
int handle_client_fin(STATE_MON *state, unsigned char *msg, unsigned int len, int direction){

	if ( state->last_message == C_CCS &&
			((state->role == CLIENT && direction == SEND) || (state->role == SERVER && direction == RECEIVE))) {
		//@ assert sentByClient(state, direction);
		//@ assert StateAfterClientCCS{Pre}(state) || StateAfterClientCCSLastMsg{Pre}(state);

		appendMessageToLog(state, msg, len);

		state->last_message = C_Fin;


		//@ assert \at(state->last_message,Here) == C_Fin;
		//@ assert HaveSameStateValuesAndRole{Here,Pre}(state);

		//@ assert \at(state->log_length,Here) == \at(state->log_length,Pre) + len ;
		//@ assert (\forall integer i; 0 <= i < \at(state->log_length,Pre) ==> \at(state->log[i],Pre) == \at(state->log[i],Here)) ;
		//@ assert(\forall integer i; 0 <= i < len ==> \at(state->log[i+\at(state->log_length,Pre)],Here) == \at(msg[i],Here)) ;

		// Takes about 2min to Alt-ergo to prove MessageAddedToLog..., z3 still not done after 5min
		//@ assert MessageAddedToLog{Here,Pre}(state, msg, len);
		//@ assert isClientFin{Here}(state, msg, len, direction, handshake);
		//@ assert ( StateAfterClientCCS{Pre}(state) || StateAfterClientCCSLastMsg{Pre}(state) );

		/*@ assert ( NextStateAfterClientFinLastMsg{Here,Pre}(state, msg, len, direction) )
					 ||
					 ( NextStateAfterClientFin{Here,Pre}(state, msg, len, direction) );
		 */
		//@ assert StateAfterClientFin{Here}(state) || StateAfterClientFinLastMsg{Here}(state);
		return ACCEPT;
	} else {
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : Rejected because the state does not match what expected.\n");
#endif
		return REJECT;
	}
}

/*@
 requires \valid(state);

 ensures StateAfterInitialState{Here}(state);
 */
void ssl3_initialize_state_mon(STATE_MON *state){
	state->role = UNDEFINED_ROLE;
	state->version = UNDEFINED_PV;
	state->kem = UNDEFINED_KEM;
	state->client_auth = UNDEFINED_AUTH;
	state->resumption = UNDEFINED;
	state->last_message = UNDEFINED_TYPE;
	state->log_length = 0;
	state->renegotiation = UNDEFINED;
	state->ntick = UNDEFINED;
	state->state_mon_initialized = YES;
}

/*
 * Main state monitor function.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;
 ensures ( \result == ACCEPT ==> isValidState(state) );
 */
int ssl3_next_message(STATE_MON *state, unsigned char* msg, unsigned int len,
		int direction, unsigned char content_type) {
	int res;

	switch (getMessageType(state, msg, len, direction, content_type)) {
	case C_Hello:
#ifdef DEBUG
		fprintf(stdout, "STATE MONITOR : Got ClientHello.\n");
#endif
			res = handle_client_hello(state, msg, len, direction);
			//@ assert ( res == ACCEPT || res == REJECT );
			//@ assert ( res == ACCEPT ) ==> StateAfterClientHello(state);
			//@ assert StateAfterClientHello(state) ==> isValidState(state);
			//@ assert ( res == ACCEPT && isValidState(state) ) || (res == REJECT ) ;
		return res;
	case S_Hello:
#ifdef DEBUG
		fprintf(stdout, "STATE MONITOR : Got ServerHello.\n");
#endif

			res = handle_server_hello(state, msg, len, direction);
			//@ assert ( res == ACCEPT || res == REJECT );
			//@ assert ( res == ACCEPT ) ==> StateAfterServerHello(state);
			//@ assert StateAfterServerHello(state) ==> isValidState(state);
			//@ assert ( res == ACCEPT && isValidState{Here}(state) ) || (res == REJECT ) ;
		return res;
	case S_Cert:
#ifdef DEBUG
		fprintf(stdout, "STATE MONITOR : Got ServerCertificate.\n");
#endif
			res = handle_server_certificate(state, msg, len, direction);
			//@ assert ( res == ACCEPT || res == REJECT );
			//@ assert ( res == ACCEPT ) ==> StateAfterServerCertificate(state);
			//@ assert StateAfterServerCertificate(state) ==> isValidState(state);
			//@ assert ( res == ACCEPT && isValidState(state) ) || (res == REJECT ) ;
		return res;
	case S_KE:
#ifdef DEBUG
		fprintf(stdout, "STATE MONITOR : Got ServerKeyExchange.\n");
#endif
			res = handle_server_key_exchange(state, msg, len, direction);
			//@ assert ( res == ACCEPT || res == REJECT );
			//@ assert ( res == ACCEPT ) ==> StateAfterServerKeyExchange(state);
			//@ assert StateAfterServerKeyExchange(state) ==> isValidState(state);
			//@ assert ( res == ACCEPT && isValidState(state) ) || (res == REJECT ) ;
		return res;
	case S_CReq:
#ifdef DEBUG
		fprintf(stdout, "STATE MONITOR : Got ServerCertificateRequest.\n");
#endif
			res = handle_server_certificate_request(state, msg, len, direction);
			//@ assert ( res == ACCEPT || res == REJECT );
			//@ assert ( res == ACCEPT ) ==> StateAfterServerCertificateRequest(state);
			//@ assert StateAfterServerCertificateRequest(state) ==> isValidState(state);
			//@ assert ( res == ACCEPT && isValidState(state) ) || (res == REJECT ) ;
		return res;
	case S_HD:
#ifdef DEBUG
		fprintf(stdout, "STATE MONITOR : Got ServerHelloDone.\n");
#endif
			res = handle_server_hello_done(state, msg, len, direction);
			//@ assert ( res == ACCEPT || res == REJECT );
			//@ assert ( res == ACCEPT ) ==> StateAfterServerHelloDone(state);
			//@ assert StateAfterServerHelloDone(state) ==> isValidState(state);
			//@ assert ( res == ACCEPT && isValidState(state) ) || (res == REJECT ) ;
		return res;
	case C_Cert:
#ifdef DEBUG
		fprintf(stdout, "STATE MONITOR : Got ClientCertificate.\n");
#endif
			res = handle_client_certificate(state, msg, len, direction);
			//@ assert ( res == ACCEPT || res == REJECT );
			//@ assert ( res == ACCEPT ) ==> StateAfterClientCertificate(state);
			//@ assert StateAfterClientCertificate(state) ==> isValidState(state);
		return res;
	case C_KE:
#ifdef DEBUG
		fprintf(stdout, "STATE MONITOR : Got ClientKeyExchange.\n");
#endif
			res = handle_client_key_exchange(state, msg, len, direction);
			//@ assert ( res == ACCEPT || res == REJECT );
			//@ assert ( res == ACCEPT ) ==> StateAfterClientKeyExchange(state);
			//@ assert StateAfterClientKeyExchange(state) ==> isValidState(state);
		return res;
	case C_CV:
#ifdef DEBUG
		fprintf(stdout, "STATE MONITOR : Got ClientCertificateVerify.\n");
#endif
			res = handle_client_certificate_verify(state, msg, len, direction);
			//@ assert ( res == ACCEPT || res == REJECT );
			//@ assert ( res == ACCEPT ) ==> StateAfterClientCertificateVerify(state);
			//@ assert StateAfterClientCertificateVerify(state) ==> isValidState(state);
		return res;
	case ServerNewSessionTicket:
#ifdef DEBUG
		fprintf(stdout, "STATE MONITOR : Got ServerNewSessionTicket.\n");
#endif
			res = handle_server_new_session_ticket(state, msg, len, direction);
			//@ assert ( res == ACCEPT || res == REJECT );
			//@ assert ( res == ACCEPT ) ==> StateAfterServerNewSessionTicket(state);
			//@ assert StateAfterServerNewSessionTicket(state) ==> isValidState(state);
		return res;
	case S_CCS:
#ifdef DEBUG
		fprintf(stdout, "STATE MONITOR : Got ServerCCS.\n");
#endif
			res = handle_server_change_cipher_specs(state, msg, len, direction);
			//@ assert ( res == ACCEPT || res == REJECT );
			//@ assert ( res == ACCEPT ) ==> StateAfterServerCCS(state);
			//@ assert StateAfterServerCCS(state) ==> isValidState(state);
		return res;
	case S_Fin:
#ifdef DEBUG
		fprintf(stdout, "STATE MONITOR : Got ServerFinished.\n");
#endif
			res = handle_server_fin(state, msg, len, direction);
			//@ assert ( res == ACCEPT || res == REJECT );
			//@ assert ( res == ACCEPT ) ==> StateAfterServerFin(state);
			//@ assert StateAfterServerFin(state) ==> isValidState(state);
		return res;
	case C_CCS:
#ifdef DEBUG
		fprintf(stdout, "STATE MONITOR : Got ClientCCS.\n");
#endif
			res = handle_client_change_cipher_specs(state, msg, len, direction);
			//@ assert ( res == ACCEPT || res == REJECT );
			//@ assert ( res == ACCEPT ) ==> ( StateAfterClientCCS(state) || StateAfterClientCCSLastMsg(state) );
			//@ assert ( StateAfterClientCCS(state) || StateAfterClientCCSLastMsg(state) ) ==> isValidState(state);
		return res;
	case C_Fin:
#ifdef DEBUG
		fprintf(stdout, "STATE MONITOR : Got ClientFinished.\n");
#endif
			res = handle_client_fin(state, msg, len, direction);
			//@ assert ( res == ACCEPT || res == REJECT );
			//@ assert ( res == ACCEPT ) ==> ( StateAfterClientFin(state) || StateAfterClientFinLastMsg(state));
			//@ assert ( StateAfterClientFin(state) || StateAfterClientFinLastMsg(state)) ==> isValidState(state);
		return res;
	default:
		res = REJECT;
		return res;
	}
#ifdef DEBUG
	fprintf(stdout,
			"STATE MONITOR : Got an unknown data type : 0x%x with content type 0x%x.\n",
			msg[0], content_type);
#endif
	res = REJECT;
	//@ assert res == REJECT;
	return res;
}


int ssl3_handle_next_message(STATE_MON *state, unsigned char* msg,
		unsigned int len, int direction, unsigned char content_type) {
	int fragment_length, new_offset, res;

	switch (content_type) {
	case alert:
		fprintf(stderr, "STATE MONITOR : got alert message.\n");
		res = ssl3_handle_next_message(state, msg, len, direction, content_type);
		break;
	case change_cipher_spec:
		fprintf(stderr, "STATE MONITOR : got CCS message.\n");
		res = ssl3_handle_next_message(state, msg, len, direction, content_type);
		break;
	case handshake:
		new_offset = 0;
		res = ACCEPT;
		do {
			fragment_length = msg[new_offset + 1] * 65536
					+ msg[new_offset + 2] * 256 + msg[new_offset + 3];

			if (msg[new_offset + 0] == hello_request) {
				if (msg[new_offset + 1] == 0 && msg[new_offset + 2] == 0
						&& msg[new_offset + 3] == 0) {
					fprintf(stderr,
							"STATE MONITOR : skipped a hello request message.\n");
					res &= ACCEPT;
				} else {
					fprintf(stderr,
							"STATE MONITOR : ERROR : non zero length for a hello request message : %d\n",
							fragment_length);
					res &= REJECT;
				}
			} else {
				res &= ssl3_handle_next_message(state, msg + new_offset,
						fragment_length+4, direction, content_type);
			}
			new_offset += fragment_length + 4;
		} while (new_offset < len);
		if ( new_offset != len ) {
		  fprintf(stderr, "STATE MONITOR : ERROR, unexpected length value");
		  return REJECT;
		  }
		break;
	default:
		fprintf(stderr, "STATE MONITOR : FAILURE, unexpected data message during handshake\n");
		res = REJECT;
		break;
	}
	return res;
}
