#include "ssl.h"
#include "s3_state_mon_common_logic.h"
#include "s3_state_mon_parsing.h"

Msg_type getMessageType(STATE_MON *state, unsigned char *msg, unsigned int len, int direction, unsigned char contentType){
	unsigned int msg_length;

	switch (contentType) {
	case change_cipher_spec :
		if ( (len != 1) || (msg[0] != 0x01) ){
			//@ assert isUnknownMessage(state, msg, len, direction, contentType);
			return UNDEFINED_TYPE;
		}
		else if (	(state->role == SERVER && direction == SEND ) ||
				(state->role == CLIENT && direction == RECEIVE) ) {
			//@ assert isServerChangeCipherSpecs(state, msg, len, direction, contentType);
			return S_CCS;
		}
		else if ((state->role == CLIENT && direction == SEND ) ||
				 (state->role == SERVER && direction == RECEIVE) ) {
			//@ assert isClientChangeCipherSpecs(state, msg, len, direction, contentType);
			return C_CCS;
		} else {
			//@ assert isUnknownMessage(state, msg, len, direction, contentType);
		}
		break;
	case alert :
		break;
	case handshake :
		if (len < 4)
			return UNDEFINED_TYPE;
		msg_length = msg[1] * 65536 + msg[2] * 256 + msg[3];
		if ( len - 4 != msg_length){
			//@ assert isUnknownMessage(state, msg, len, direction, contentType);
			return UNDEFINED_TYPE;
		}
		switch (msg[0]){
		case C_Hello:
			//@ assert isClientHello(msg, len, contentType);
			return C_Hello;
		case S_Hello:
			//@ assert isServerHello(msg, len, contentType);
			return S_Hello;
		case certificate:
			if ((state->role == SERVER && direction == SEND ) ||
							(state->role == CLIENT && direction == RECEIVE)) {
				//@ assert isServerCertificate(state, msg, len, direction, contentType);
				return S_Cert;
			} else if ((state->role == CLIENT && direction == SEND)
					|| (state->role == SERVER && direction == RECEIVE)) {
				//@ assert isClientCertificate(state, msg, len, direction, contentType);
				return C_Cert;
			} else {
				//@ assert isUnknownMessage(state, msg, len, direction, contentType);
				return UNDEFINED_TYPE;
			}
			break;
		case S_KE:
			//@ assert isServerKeyExchange(msg, len, contentType);
			return S_KE;
		case S_CReq:
			//@ assert isServerCertificateRequest(msg, len, contentType);
			return S_CReq;
		case S_HD:
			//@ assert isServerHelloDone(msg, len, contentType);
			return S_HD;
		case C_KE:
			//@ assert isClientKeyExchange(msg, len, contentType);
			return C_KE;
		case C_CV:
			//@ assert isClientCertificateVerify(msg, len, contentType);
			return C_CV;
		case finished:
			if ((state->role == SERVER && direction == SEND)
					|| (state->role == CLIENT && direction == RECEIVE)) {
				//@ assert isServerFin(state, msg, len, direction, contentType);
				return S_Fin;
			} else if ((state->role == CLIENT && direction == SEND)
					|| (state->role == SERVER && direction == RECEIVE)) {
				//@ assert isClientFin(state, msg, len, direction, contentType);
				return C_Fin;
			} else {
				//@ assert isUnknownMessage(state, msg, len, direction, contentType);
				return UNDEFINED_TYPE;
			}
			break;
		case ServerNewSessionTicket:
			//@ assert isServerNewSessionTicket(msg, len, contentType);
			return ServerNewSessionTicket;
		}
		break;
	}
	//@ assert isUnknownMessage(state, msg, len, direction, contentType);
	return UNDEFINED_TYPE;
}



PV getPV(unsigned char *msg, unsigned int len){
	if (len < 6){
		//@ assert isUnknownPV(msg, len);
		return UNDEFINED_PV;
	}
	if(msg[4] != 0x03){
		//@ assert isUnknownPV(msg, len);
		return UNDEFINED_PV;
	}
	else {
		switch (msg[5]){
		case 0x00:
			//@ assert isSSL3(msg, len);
			return SSL3;
		case 0x01 :
			//@ assert isTLS1_0(msg, len);
			return TLS1;
		case 0x02:
			//@ assert isTLS1_1(msg, len);
			return TLS1_1;
		case 0x03:
			//@ assert isTLS1_2(msg, len);
			return TLS1_2;
		default :
			//@ assert isUnknownPV(msg, len);
			return UNDEFINED_PV;
		}
	}
}


KEM getKEM(unsigned char *msg, unsigned int len){
	/* 38 =[ServerHello] HandshakeType(1) + length(3) + Version(2) + Random(32) */
	if (len < 39){
		//@ assert msg[38] >= 0;
		//@ assert isUnknownCS(msg, len);
		return UNDEFINED_KEM;
	}
	int offset = 38;
	offset += 1 + msg[offset]; // Length of the session id
	if (len < offset+2){
		//@ assert isUnknownCS(msg, len);
		return UNDEFINED_KEM;
	}
	switch (msg[offset]) {
	case 0x00:
		switch (msg[offset+1]){
		case 0x01:
		case 0x02:
		case 0x3B:
		case 0x04:
		case 0x05:
		case 0x07:
		case 0x09:
		case 0x0A:
		case 0x2F:
		case 0x35:
		case 0x3C:
		case 0x3D:
		case 0x84:
		case 0x96:
		case 0x9C:
		case 0x9D:
		case 0xBA:
		case 0xC0:
			//@ assert isRSA(msg, len);
			return CS_RSA;
		case 0x03:
		case 0x06:
		case 0x08:
			//@ assert isRSAExport(msg, len);
			return RSAExport;
		case 0x0C:
		case 0x0D:
		case 0x0F:
		case 0x10:
		case 0x30:
		case 0x31:
		case 0x36:
		case 0x37:
		case 0x3E:
		case 0x3F:
		case 0x42:
		case 0x43:
		case 0x68:
		case 0x69:
		case 0x85:
		case 0x86:
		case 0x97:
		case 0x98:
		case 0xA0:
		case 0xA1:
		case 0xA4:
		case 0xA5:
		case 0xBB:
		case 0xBC:
		case 0xC1:
		case 0xC2:
			//@ assert isDH(msg, len);
			return DiffieHellman;
		case 0x12:
		case 0x13:
		case 0x15:
		case 0x16:
		case 0x32:
		case 0x33:
		case 0x38:
		case 0x39:
		case 0x40:
		case 0x44:
		case 0x45:
		case 0x67:
		case 0x6A:
		case 0x6B:
		case 0x87:
		case 0x88:
		case 0x99:
		case 0x9A:
		case 0x9E:
		case 0x9F:
		case 0xA2:
		case 0xA3:
		case 0xBD:
		case 0xBE:
		case 0xC3:
		case 0xC4:
			//@ assert isDHE(msg, len);
			return DHE;
		case 0x11:
		case 0x14:
			//@ assert isDHEExport(msg, len);
			return DHEExport;
		case 0x18:
		case 0x1A:
		case 0x1B:
		case 0x34:
		case 0x3A:
		case 0x46:
		case 0x6C:
		case 0x6D:
		case 0x89:
		case 0x9B:
		case 0xA6:
		case 0xA7:
		case 0xBF:
		case 0xC5:
			//@ assert isDHanon(msg, len);
			return DHanon;
		default:
			//@ assert isUnknownCS(msg, len);
			return UNDEFINED_KEM;
		}
		break;

	case 0xC0:
		switch(msg[offset+1]){
		case 0x3C:
		case 0x3D:
		case 0x50:
		case 0x51:
		case 0x7A:
		case 0x7B:
		case 0x9C:
		case 0x9D:
		case 0xA0:
		case 0xA1:
			//@ assert isRSA(msg, len);
			return CS_RSA;
		case 0x3E:
		case 0x3F:
		case 0x40:
		case 0x41:
		case 0x54:
		case 0x55:
		case 0x58:
		case 0x59:
		case 0x7E:
		case 0x7F:
		case 0x82:
		case 0x83:
			//@ assert isDH(msg, len);
			return DiffieHellman;
		case 0x46:
		case 0x47:
		case 0x5A:
		case 0x5B:
		case 0x84:
		case 0x85:
			//@ assert isDHanon(msg, len);
			return DHanon;
		case 0x42:
		case 0x43:
		case 0x44:
		case 0x45:
		case 0x52:
		case 0x53:
		case 0x56:
		case 0x57:
		case 0x7C:
		case 0x7D:
		case 0x80:
		case 0x81:
		case 0x9E:
		case 0x9F:
		case 0xA2:
		case 0xA3:
			//@ assert isDHE(msg, len);
			return DHE;
		case 0x01:
		case 0x02:
		case 0x03:
		case 0x04:
		case 0x05:
		case 0x0B:
		case 0x0C:
		case 0x0D:
		case 0x0E:
		case 0x0F:
		case 0x25:
		case 0x26:
		case 0x29:
		case 0x2A:
		case 0x2D:
		case 0x2E:
		case 0x31:
		case 0x32:
		case 0x4A:
		case 0x4B:
		case 0x4E:
		case 0x4F:
		case 0x5E:
		case 0x5F:
		case 0x62:
		case 0x63:
		case 0x74:
		case 0x75:
		case 0x78:
		case 0x79:
		case 0x88:
		case 0x89:
		case 0x8C:
		case 0x8D:
			//@ assert isECDH(msg, len);
			return ECDH;
		case 0x06:
		case 0x07:
		case 0x08:
		case 0x09:
		case 0x0A:
		case 0x10:
		case 0x11:
		case 0x12:
		case 0x13:
		case 0x14:
		case 0x23:
		case 0x24:
		case 0x27:
		case 0x28:
		case 0x2B:
		case 0x2C:
		case 0x2F:
		case 0x30:
		case 0x48:
		case 0x49:
		case 0x4C:
		case 0x4D:
		case 0x5C:
		case 0x5D:
		case 0x60:
		case 0x61:
		case 0x72:
		case 0x73:
		case 0x76:
		case 0x77:
		case 0x8A:
		case 0x8B:
		case 0xAc:
		case 0xAD:
		case 0xAE:
		case 0xAF:
			//@ assert isECDHE(msg, len);
			return ECDHE;
		case 0x15:
		case 0x16:
		case 0x17:
		case 0x18:
		case 0x19:
			//@ assert isECDHanon(msg, len);
			return ECDHanon;
		default :
			//@ assert isUnknownCS(msg, len);
			return UNDEFINED_KEM;
		}
		break;
	default:
		//@ assert isUnknownCS(msg, len);
		return UNDEFINED_KEM;
	}
}


Auth checkClientCertificate(STATE_MON *state, unsigned char *msg, unsigned int len, int direction){
	if (msg[1] == 0x00 && msg[2] == 0x00 && msg[3] == 0x00)
		return AUTH_NOT_PROVIDED;
	else
		return AUTH_PROVIDED;
}
