#include <limits.h>
#include <string.h>
#include <stdio.h>
#include "ssl_locl.h"
#include "ssl.h"
#include "s3_state_mon_common_logic.h"
#include "s3_state_mon.h"
#include <openssl/buffer.h>
#include <openssl/rand.h>
#include <openssl/objects.h>
#include <openssl/evp.h>
#include <openssl/x509.h>

#define DEBUG

/*
 * Wrapper for the memcpy function in order to give it some logic.
 */
unsigned char *memcpy_uchar(unsigned char *dest, unsigned char const *src,
		unsigned int n) {
	return (unsigned char*) memcpy((unsigned char*) dest,
			(unsigned char* const ) src, n);
}

/*
 * Parsing function for the message type, the complete verification and separation is done in s3_state_mon_parsing.c
 * and s3_state_mon_parsing.h
 */
/*@
 assigns \nothing;

 ensures (isClientHello(msg, len, content_type) 								<==> \result == C_Hello) 				&&
 (isServerHello(msg, len, content_type) 								<==> \result == S_Hello )				&&
 (isServerCertificate(state, msg, len, direction, content_type) 		<==> \result == S_Cert) 				&&
 (isServerKeyExchange(msg, len, content_type) 							<==> \result == S_KE) 					&&
 (isServerCertificateRequest(msg, len, content_type) 					<==> \result == S_CReq) 				&&
 (isServerHelloDone(msg, len, content_type) 							<==> \result == S_HD )					&&
 (isClientCertificate(state, msg, len, direction, content_type) 		<==> \result == C_Cert) 				&&
 (isClientKeyExchange(msg, len, content_type) 							<==> \result == C_KE) 					&&
 (isClientCertificateVerify(msg, len, content_type) 					<==> \result == C_CV )					&&
 (isServerChangeCipherSpecs(state, msg, len, direction, content_type)	<==> \result ==	S_CCS) 					&&
 (isServerFin(state, msg, len, direction, content_type) 				<==> \result == S_Fin) 					&&
 (isClientChangeCipherSpecs(state, msg, len, direction, content_type)	<==> \result == C_CCS )					&&
 (isClientFin(state, msg, len, direction, content_type) 				<==> \result == C_Fin) 					&&
 (isServerNewSessionTicket(msg, len, content_type) 						<==> \result == ServerNewSessionTicket) &&
 (isUnknownMessage(state, msg, len, direction, content_type)			<==> \result == UNDEFINED_TYPE )		;
 */
Msg_type getMessageType(STATE_MON *state, unsigned char *msg, unsigned int len,
		int direction, unsigned char content_type) {
	unsigned int msg_length;

	switch (content_type) {
	case change_cipher_spec:
		if ((len != 1) || (msg[0] != 0x01)) {
#ifdef DEBUG
			fprintf(stderr,
					"STATE MONITOR : ERROR : wrong CCS message format, got len = %d and msg[0] = 0x%x instead of 1 and 0x01.\n",
					len, msg[0]);
#endif
			//@ assert isUnknownMessage(state, msg, len, direction, content_type);
			return UNDEFINED_TYPE;
		} else if ((state->role == SERVER && direction == SEND)
				|| (state->role == CLIENT && direction == RECEIVE)) {
			//@ assert isServerChangeCipherSpecs(state, msg, len, direction, content_type);
			return S_CCS;
		} else if ((state->role == CLIENT && direction == SEND)
				|| (state->role == SERVER && direction == RECEIVE)) {
			//@ assert isClientChangeCipherSpecs(state, msg, len, direction, content_type);
			return C_CCS;
		} else {
			//@ assert isUnknownMessage(state, msg, len, direction, content_type);
#ifdef DEBUG
			fprintf(stderr,
					"STATE MONITOR : Error, my role is %s and the direction is %s.\n",
					state->role == CLIENT ? "CLIENT" : "SERVER",
					direction == SEND ? "SEND" : "RECEIVE");
#endif
		}
		break;
	case alert:
		fprintf(stderr, "STATE MONITOR : ERROR : got an alert message, handshake should be aborted.\n");
		break;
	case application_data:
		fprintf(stderr, "STATE MONITOR : ERROR : got application data, handshake should be aborted.\n");
	case handshake:
		if (len < 4){
			fprintf(stderr, "STATE MONITOR : ERROR : ran into message of negative length.\n");
			return UNDEFINED_TYPE;
		}
		msg_length = msg[1] * 65536 + msg[2] * 256 + msg[3];
		if ( !( (len-4) == msg_length)  ) {
#ifdef DEBUG
			fprintf(stderr,
					"STATE MONITOR : ERROR : wrong handshake message format, expected len - 4 == msg_length but len == %d and msg_length == %d.\n",
					len, msg_length);
#endif
			//@ assert isUnknownMessage(state, msg, len, direction, content_type);
			return UNDEFINED_TYPE;
		}
		switch (msg[0]) {
		case C_Hello:
			//@ assert isClientHello(msg, len, content_type);
			return C_Hello;
		case S_Hello:
			//@ assert isServerHello(msg, len, content_type);
			return S_Hello;
		case certificate:
			if ((state->role == SERVER && direction == SEND)
					|| (state->role == CLIENT && direction == RECEIVE)) {
				//@ assert isServerCertificate(state, msg, len, direction, content_type);
				return S_Cert;
			} else if ((state->role == CLIENT && direction == SEND)
					|| (state->role == SERVER && direction == RECEIVE)) {
				//@ assert isClientCertificate(state, msg, len, direction, content_type);
				return C_Cert;
			} else {
				//@ assert isUnknownMessage(state, msg, len, direction, content_type);
				return UNDEFINED_TYPE;
			}
			break;
		case S_KE:
			//@ assert isServerKeyExchange(msg, len, content_type);
			return S_KE;
		case S_CReq:
			//@ assert isServerCertificateRequest(msg, len, content_type);
			return S_CReq;
		case S_HD:
			//@ assert isServerHelloDone(msg, len, content_type);
			return S_HD;
		case C_KE:
			//@ assert isClientKeyExchange(msg, len, content_type);
			return C_KE;
		case C_CV:
			//@ assert isClientCertificateVerify(msg, len, content_type);
			return C_CV;
		case finished:
			if ((state->role == SERVER && direction == SEND)
					|| (state->role == CLIENT && direction == RECEIVE)) {
				//@ assert isServerFin(state, msg, len, direction, content_type);
				return S_Fin;
			} else if ((state->role == CLIENT && direction == SEND)
					|| (state->role == SERVER && direction == RECEIVE)) {
				//@ assert isClientFin(state, msg, len, direction, content_type);
				return C_Fin;
			} else {
				//@ assert isUnknownMessage(state, msg, len, direction, content_type);
				return UNDEFINED_TYPE;
			}
			break;
		case ServerNewSessionTicket:
			//@ assert isServerNewSessionTicket(msg, len, content_type);
			return ServerNewSessionTicket;
		default:
			fprintf(stderr, "STATE MONITOR : ERROR : got an unknown handshake type : %d\n", msg[0]);
			return UNDEFINED_TYPE;
		}
		break;
	}
	//@ assert isUnknownMessage(state, msg, len, direction, content_type);
	return UNDEFINED_TYPE;
}

/*
 * Parsing function for the protocol version, the complete verification and separation is done in s3_state_mon_parsing.c
 * and s3_state_mon_parsing.h
 */
/*@
 requires isMessageLength(msg, len);

 assigns \nothing;

 ensures
 (\result == SSL3 <==> isSSL3(msg, len)) &&
 (\result == TLS1 <==> isTLS1_0(msg, len)) &&
 (\result == TLS1_1 <==> isTLS1_1(msg, len)) &&
 (\result == TLS1_2 <==> isTLS1_2(msg, len)) &&
 (\result == UNDEFINED_PV <==> isUnknownPV(msg, len));
 */
PV getPV(unsigned char *msg, unsigned int len) {
	if (len < 6) {
#ifdef DEBUG
		//fprintf(stderr, "STATE MONITOR : ERROR : Server Hello length was too small to parse the protocol version : %d bytes.\n", len);
#endif
		//@ assert isUnknownPV(msg, len);
		return UNDEFINED_PV;
	}
	if (msg[4] != 0x03) {
#ifdef DEBUG
		fprintf(stderr,
				"STATE MONITOR : Unknown protocol version : 0x%x 0x%x\n",
				msg[4], msg[5]);
#endif
		//@ assert isUnknownPV(msg, len);
		return UNDEFINED_PV;
	} else {
		switch (msg[5]) {
		case 0x00:
			fprintf(stderr, "STATE MONITOR : protocol version is SSL3.\n");
			//@ assert isSSL3(msg, len);
			return SSL3;
		case 0x01:
			fprintf(stderr, "STATE MONITOR : protocol version is TLS1.0.\n");
			//@ assert isTLS1_0(msg, len);
			return TLS1;
		case 0x02:
			fprintf(stderr, "STATE MONITOR : protocol version is TLS1.1.\n");
			//@ assert isTLS1_1(msg, len);
			return TLS1_1;
		case 0x03:
			fprintf(stderr, "STATE MONITOR : protocol version is TLS1.2.\n");
			//@ assert isTLS1_2(msg, len);
			return TLS1_2;
		default:
#ifdef DEBUG
			fprintf(stderr,
					"STATE MONITOR : Unknown protocol version : 0x%x 0x%x\n",
					msg[4], msg[5]);
#endif
			//@ assert isUnknownPV(msg, len);
			return UNDEFINED_PV;
		}
	}
}

/*
 * Parsing function for the key exchange method, the complete verification and separation is done in s3_state_mon_parsing.c
 * and s3_state_mon_parsing.h
 */
/*@
 requires isMessageLength(msg, len);

 assigns \nothing;

 ensures
 (\result == DiffieHellman 	<==> isDH(msg, len))	  		&&
 (\result == DHE 			<==> isDHE(msg, len)) 			&&
 (\result == DHEExport		<==> isDHEExport(msg, len))		&&
 (\result == DHanon 		<==> isDHanon(msg, len))		&&
 (\result == ECDH 			<==> isECDH(msg, len))  		&&
 (\result == ECDHE 			<==> isECDHE(msg, len)) 		&&
 (\result == ECDHanon 		<==> isECDHanon(msg, len)) 		&&
 (\result == CS_RSA 		<==> isRSA(msg, len)) 			&&
 (\result == RSAExport 		<==> isRSAExport(msg, len))  	&&
 (\result == UNDEFINED_KEM 	<==> isUnknownKEM(msg, len)) 	 ;
 */
KEM getKEM(unsigned char *msg, unsigned int len) {
	/* 38 =[ServerHello] HandshakeType(1) + length(3) + Version(2) + Random(32) */
	if (len < 39) {
#ifdef DEBUG
		//fprintf(stderr, "STATE MONITOR : ERROR : Server Hello length was too small to parse the key exchange method : %d bytes.\n", len);
#endif
		//@ assert msg[38] >= 0;
		//@ assert isUnknownKEM(msg, len);
		return UNDEFINED_KEM;
	}
	int offset = 38;
	offset += 1 + msg[offset]; // Length of the session id
	if (len < offset + 2) {
		//@ assert isUnknownKEM(msg, len);
		return UNDEFINED_KEM;
	}
	switch (msg[offset]) {
	case 0x00:
		switch (msg[offset + 1]) {
		case 0x01:
		case 0x02:
		case 0x3B:
		case 0x04:
		case 0x05:
		case 0x07:
		case 0x09:
		case 0x0A:
		case 0x2F:
		case 0x35:
		case 0x3C:
		case 0x3D:
		case 0x84:
		case 0x96:
		case 0x9C:
		case 0x9D:
		case 0xBA:
		case 0xC0:
#ifdef DEBUG
			//fprintf(stderr, "STATE MONITOR : Got an RSA cipher suite.\n");
#endif
			//@ assert isRSA(msg, len);
			return CS_RSA;
		case 0x03:
		case 0x06:
		case 0x08:
#ifdef DEBUG
			//fprintf(stderr, "STATE MONITOR : Got an RSA EXPORT cipher suite.\n");
#endif
			//@ assert isRSAExport(msg, len);
			return RSAExport;
		case 0x0C:
		case 0x0D:
		case 0x0F:
		case 0x10:
		case 0x30:
		case 0x31:
		case 0x36:
		case 0x37:
		case 0x3E:
		case 0x3F:
		case 0x42:
		case 0x43:
		case 0x68:
		case 0x69:
		case 0x85:
		case 0x86:
		case 0x97:
		case 0x98:
		case 0xA0:
		case 0xA1:
		case 0xA4:
		case 0xA5:
		case 0xBB:
		case 0xBC:
		case 0xC1:
		case 0xC2:
#ifdef DEBUG
			//fprintf(stderr, "STATE MONITOR : Got an DH cipher suite.\n");
#endif
			//@ assert isDH(msg, len);
			return DiffieHellman;
		case 0x12:
		case 0x13:
		case 0x15:
		case 0x16:
		case 0x32:
		case 0x33:
		case 0x38:
		case 0x39:
		case 0x40:
		case 0x44:
		case 0x45:
		case 0x67:
		case 0x6A:
		case 0x6B:
		case 0x87:
		case 0x88:
		case 0x99:
		case 0x9A:
		case 0x9E:
		case 0x9F:
		case 0xA2:
		case 0xA3:
		case 0xBD:
		case 0xBE:
		case 0xC3:
		case 0xC4:
#ifdef DEBUG
			//fprintf(stderr, "STATE MONITOR : Got an DHE cipher suite.\n");
#endif
			//@ assert isDHE(msg, len);
			return DHE;
		case 0x11:
		case 0x14:
#ifdef DEBUG
			//fprintf(stderr, "STATE MONITOR : Got an DHE EXPORT cipher suite.\n");
#endif
			//@ assert isDHEExport(msg, len);
			return DHEExport;
		case 0x18:
		case 0x1A:
		case 0x1B:
		case 0x34:
		case 0x3A:
		case 0x46:
		case 0x6C:
		case 0x6D:
		case 0x89:
		case 0x9B:
		case 0xA6:
		case 0xA7:
		case 0xBF:
		case 0xC5:
#ifdef DEBUG
			//fprintf(stderr, "STATE MONITOR : Got an DH anonymous cipher suite.\n");
#endif
			//@ assert isDHanon(msg, len);
			return DHanon;
		default:
#ifdef DEBUG
			//fprintf(stderr, "STATE MONITOR : Unknown cipher suite : 0x00 0x%x\n", msg[offset+1]);
#endif
			//@ assert isUnknownKEM(msg, len);
			return UNDEFINED_KEM;
		}
		break;

	case 0xC0:
		switch (msg[offset + 1]) {
		case 0x3C:
		case 0x3D:
		case 0x50:
		case 0x51:
		case 0x7A:
		case 0x7B:
		case 0x9C:
		case 0x9D:
		case 0xA0:
		case 0xA1:
#ifdef DEBUG
			//fprintf(stderr, "STATE MONITOR : Got an RSA cipher suite.\n");
#endif
			//@ assert isRSA(msg, len);
			return CS_RSA;
		case 0x3E:
		case 0x3F:
		case 0x40:
		case 0x41:
		case 0x54:
		case 0x55:
		case 0x58:
		case 0x59:
		case 0x7E:
		case 0x7F:
		case 0x82:
		case 0x83:
#ifdef DEBUG
			//fprintf(stderr, "STATE MONITOR : Got an DH cipher suite.\n");
#endif
			//@ assert isDH(msg, len);
			return DiffieHellman;
		case 0x46:
		case 0x47:
		case 0x5A:
		case 0x5B:
		case 0x84:
		case 0x85:
#ifdef DEBUG
			//fprintf(stderr, "STATE MONITOR : Got an DH anonymous cipher suite.\n");
#endif
			//@ assert isDHanon(msg, len);
			return DHanon;
		case 0x42:
		case 0x43:
		case 0x44:
		case 0x45:
		case 0x52:
		case 0x53:
		case 0x56:
		case 0x57:
		case 0x7C:
		case 0x7D:
		case 0x80:
		case 0x81:
		case 0x9E:
		case 0x9F:
		case 0xA2:
		case 0xA3:
#ifdef DEBUG
			//fprintf(stderr, "STATE MONITOR : Got an DHE cipher suite.\n");
#endif
			//@ assert isDHE(msg, len);
			return DHE;
		case 0x01:
		case 0x02:
		case 0x03:
		case 0x04:
		case 0x05:
		case 0x0B:
		case 0x0C:
		case 0x0D:
		case 0x0E:
		case 0x0F:
		case 0x25:
		case 0x26:
		case 0x29:
		case 0x2A:
		case 0x2D:
		case 0x2E:
		case 0x31:
		case 0x32:
		case 0x4A:
		case 0x4B:
		case 0x4E:
		case 0x4F:
		case 0x5E:
		case 0x5F:
		case 0x62:
		case 0x63:
		case 0x74:
		case 0x75:
		case 0x78:
		case 0x79:
		case 0x88:
		case 0x89:
		case 0x8C:
		case 0x8D:
#ifdef DEBUG
			//fprintf(stderr, "STATE MONITOR : Got an ECDH cipher suite.\n");
#endif
			//@ assert isECDH(msg, len);
			return ECDH;
		case 0x06:
		case 0x07:
		case 0x08:
		case 0x09:
		case 0x0A:
		case 0x10:
		case 0x11:
		case 0x12:
		case 0x13:
		case 0x14:
		case 0x23:
		case 0x24:
		case 0x27:
		case 0x28:
		case 0x2B:
		case 0x2C:
		case 0x2F:
		case 0x30:
		case 0x48:
		case 0x49:
		case 0x4C:
		case 0x4D:
		case 0x5C:
		case 0x5D:
		case 0x60:
		case 0x61:
		case 0x72:
		case 0x73:
		case 0x76:
		case 0x77:
		case 0x8A:
		case 0x8B:
		case 0xAc:
		case 0xAD:
		case 0xAE:
		case 0xAF:
#ifdef DEBUG
			//fprintf(stderr, "STATE MONITOR : Got an ECDHE cipher suite.\n");
#endif
			//@ assert isECDHE(msg, len);
			return ECDHE;
		case 0x15:
		case 0x16:
		case 0x17:
		case 0x18:
		case 0x19:
#ifdef DEBUG
			//fprintf(stderr, "STATE MONITOR : Got an ECDH anonymous cipher suite.\n");
#endif
			//@ assert isECDHanon(msg, len);
			return ECDHanon;
		default:
#ifdef DEBUG
			//fprintf(stderr, "STATE MONITOR : Unknown protocol version : 0x0C 0x%x\n", msg[offset+1]);
#endif
			//@ assert isUnknownKEM(msg, len);
			return UNDEFINED_KEM;
		}
		break;
	default:
#ifdef DEBUG
		//fprintf(stderr, "STATE MONITOR : Unknown protocol version : 0x%x 0x%x\n", msg[offset], msg[offset+1]);
#endif
		//@ assert isUnknownKEM(msg, len);
		return UNDEFINED_KEM;
	}
}

/*
 * Parsing function checking whether the client certificate is empty or not.
 */
/*@
 requires isClientCertificate(state, msg, len, direction, handshake);

 assigns \nothing;

 behavior other_certificate:
 assumes (msg[1] != 0 || msg[2] != 0 || msg[3] != 0) ;
 ensures \result == AUTH_PROVIDED <==> isNonEmptyClientCertificate{Pre}(state, msg, len, direction) ;
 behavior empty_certificate :
 assumes msg[1] == 0 && msg[2] == 0 && msg[3] == 0;
 ensures \result == AUTH_NOT_PROVIDED <==> isEmptyClientCertificate{Pre}(state, msg, len, direction);
 complete behaviors other_certificate, empty_certificate;
 disjoint behaviors other_certificate, empty_certificate;
 */
Auth checkClientCertificate(STATE_MON *state, unsigned char *msg,
		unsigned int len, int direction) {
	if (msg[1] == 0x00 && msg[2] == 0x00 && msg[3] == 0x00)
		return AUTH_NOT_PROVIDED;
	else
		return AUTH_PROVIDED;
}

/*
 * Takes the current state, the last message received and its length and appends the message to the log of
 * the state monitor.
 */
/*@
 requires \valid(state) && \valid(msg+(0..len-1)) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(state->log+(0..state->log_length+len-1), state+(0..sizeof(state)-1), msg+(0..len-1));

 ensures MessageAddedToLog{Here,Pre}(state, msg, len);

 assigns state->log[\at(state->log_length,Pre)..\at(state->log_length,Pre)+len-1],
 state->log_length;
 */
void appendMessageToLog(STATE_MON *state, unsigned char *msg, unsigned int len) {
	memcpy_uchar(state->log + state->log_length, msg, len);
	state->log_length += len;
}

/*
 * Handles the ClientHello messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isClientHello(msg, len, handshake);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success :
 assumes  StateAfterInitialState(state);

 ensures \result == ACCEPT <==> StateAfterClientHello{Here}(state);

 behavior failure :
 assumes !(StateAfterInitialState(state));

 ensures \result == REJECT;

 disjoint behaviors success, failure;
 complete behaviors success, failure;
 */
int handle_client_hello(STATE_MON *state, unsigned char *msg, unsigned int len,
		int direction) {

	if (state->version == UNDEFINED_PV&&
	state->role == UNDEFINED_ROLE &&
	state->kem == UNDEFINED_KEM &&
	state->last_message == UNDEFINED_TYPE &&
	state->log_length == 0 &&
	state->client_auth == UNDEFINED_AUTH &&
	state->resumption == UNDEFINED &&
	state->renegotiation == UNDEFINED &&
	state->ntick == UNDEFINED) {

		//@ assert StateAfterInitialState{Pre}(state);

		/*@
		 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
		 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));

		 ensures MessageAddedToLog{Here,Pre}(state, msg, len);
		 ensures state->role == CLIENT || state->role == SERVER;
		 ensures state->last_message == C_Hello;

		 assigns state->log_length, state->log[\at(state->log_length,Pre)..\at(state->log_length,Pre)+len-1], state->last_message, state->role;
		 */
		{
			appendMessageToLog(state, msg, len);

			if (direction == SEND)
				state->role = CLIENT;
			else
				state->role = SERVER;

			state->last_message = C_Hello;
			//@ assert \at(state->log_length, Pre) + len == \at(state->log_length, Here);
			//@ assert \forall integer i; 0 <= i < len ==> \at(state->log[i + \at(state->log_length,Pre)],Here) == \at(msg[i],Pre);
			//@ assert \forall integer i; 0 <= i < \at(state->log_length, Pre) ==> \at(state->log[i],Here) == \at(state->log[i],Pre);

			//@ assert MessageAddedToLog{Here,Pre}(state, msg, len);
		}

		//@ assert state->last_message == C_Hello ;
		//@ assert HaveSameStateValuesButRole{Here,Pre}(state);
		//@ assert (state->role == CLIENT || state->role == SERVER );
		//@ assert MessageAddedToLog{Here,Pre}(state,msg, len);
		//@ assert isClientHello{Here}(msg, len, handshake) ;
		//@ assert StateAfterInitialState{Pre}(state);

		//@ assert NextStateAfterClientHello{Here,Pre}(state, msg, len);

		//@ assert StateAfterClientHello{Here}(state);

		return ACCEPT;

	} else {
#ifdef DEBUG
		fprintf(stderr,
				"STATE MONITOR : Rejected a non-initial state (with state_mon_initialized == %d) : \nversion : %d\nrole : %d\ncs : %d, \nlast_message : %d\n",
				state->state_mon_initialized, state->version, state->role,
				state->kem, state->last_message);
		fprintf(stderr,
				"log_length : %d\nclient_auth : %d\nresumption : %d, \nrenegociation : %d\n",
				state->log_length, state->client_auth, state->resumption,
				state->renegotiation);
#endif
		return REJECT;
	}
}

/*
 * Handles the ServerHello messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isServerHello(msg, len, handshake);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success :

 assumes StateAfterClientHello{Pre}(state)
 && sentByServer(state, direction)
 && !isUnknownPV(msg, len) && !isUnknownKEM(msg, len);

 ensures \result == ACCEPT ==> StateAfterServerHello{Here}(state);
 //assigns
 //	 state->version,
 //	 state->kem,
 //	 state->last_message,
 //	 state->log[\at(state->log_length,Pre)..(\at(state->log_length,Pre)+len-1)],
 //	 state->log_length;

 behavior failure :
 assumes !(StateAfterClientHello{Pre}(state)
 && sentByServer(state, direction)
 && !isUnknownPV(msg, len) && !isUnknownKEM(msg, len) );

 ensures \result == REJECT;
 //assigns \nothing;

 complete behaviors success, failure;
 disjoint behaviors success, failure;
 */
int handle_server_hello(STATE_MON *state, unsigned char *msg, unsigned int len,
		int direction) {

	if (state->last_message == C_Hello
			&& ((state->role == CLIENT && direction == RECEIVE)
					|| (state->role == SERVER && direction == SEND))) {
		//@ assert StateAfterClientHello{Pre}(state);

		PV version = getPV(msg, len);
		KEM kem = getKEM(msg, len);

		if (version == UNDEFINED_PV || kem == UNDEFINED_KEM) {
			if (version == UNDEFINED_PV) {
#ifdef DEBUG
				fprintf(stderr,
						"STATE MONITOR : WARNING : unsupported TLS version, failing handshake.\n");
#endif
			} else {
#ifdef DEBUG
				fprintf(stderr,
						"STATE MONITOR : WARNING : unsupported cipher suite, failing handshake.\n");
#endif
			}
#ifdef DEBUG
			int i;
			/*
			fprintf(stderr, "STATE_MONITOR : wrong message :\n");
			for (i = 0; i < len; i++) {
				fprintf(stderr, "0x%x ", msg[i]);
				if (i % 10 == 0)
					fprintf(stderr, "\n");
			}
			*/
#endif
			return REJECT;
		} else {
			//@ assert sentByServer(state, direction);

			/*@
			 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
			 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
			 requires !isUnknownPV(msg, len);
			 requires !isUnknownKEM(msg, len);

			 ensures MessageAddedToLog{Here,Pre}(state, msg, len);
			 ensures AgreeOnVersion{Here}(state, msg, len);
			 ensures AgreeOnKEM{Here}(state, msg, len);
			 ensures state->last_message == S_Hello;

			 assigns state->log_length, state->log[\at(state->log_length,Pre)..\at(state->log_length,Pre)+len-1], state->last_message, state->version, state->kem;
			 */
			{
				appendMessageToLog(state, msg, len);

				state->version = version;
				state->kem = kem;
				state->last_message = S_Hello;
				//@ assert \at(state->log_length, Pre) + len == \at(state->log_length, Here);
				//@ assert \forall integer i; 0 <= i < len ==> \at(state->log[i + \at(state->log_length,Pre)],Here) == \at(msg[i],Pre);
				//@ assert \forall integer i; 0 <= i < \at(state->log_length, Pre) ==> \at(state->log[i],Here) == \at(state->log[i],Pre);

				//@ assert MessageAddedToLog{Here,Pre}(state, msg, len);
			}
			//@ assert \at(state->last_message,Pre) == C_Hello ;
			//@ assert \at(state->last_message,Here) == S_Hello;
			//@ assert HaveSameStateValuesButVersionAndCs{Here,Pre}(state);
			//@ assert AgreeOnVersion{Here}(state, msg, len);
			//@ assert AgreeOnKEM{Here}(state, msg, len);

			//@ assert MessageAddedToLog{Here,Pre}(state,msg, len);
			//@ assert isServerHello{Here}(msg, len, handshake);
			//@ assert StateAfterClientHello{Pre}(state);

			//@ assert NextStateAfterServerHello{Here,Pre}(state, msg, len);

			//@ assert StateAfterServerHello{Here}(state);
			return ACCEPT;
		}
	} else {
#ifdef DEBUG
		fprintf(stderr,
				"STATE MONITOR : handle_server_hello rejected a message because the previous state was not appropriate to get a server hello.\n");
#endif
		return REJECT;
	}
}

/*
 * Handles the ServerCertificate messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isServerCertificate(state, msg, len, direction, handshake);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success :
 assumes sentByServer(state, direction) 	&&
 StateAfterServerHello{Pre}(state)	&&
 CsRequiresServerCertificate{Pre}(state);

 ensures \result == ACCEPT ==> StateAfterServerCertificate(state) ;

 //assigns
 //	 state->resumption,
 //	 state->last_message,
 //	 state->log_length,
 //	 state->log[\at(state->log_length,Pre)..(\at(state->log_length,Pre)+len-1)];

 behavior failure :
 assumes !( sentByServer(state, direction) 	&&
 StateAfterServerHello{Pre}(state)	&&
 CsRequiresServerCertificate{Pre}(state));

 ensures \result == REJECT;

 //assigns \nothing;
 */
int handle_server_certificate(STATE_MON *state, unsigned char *msg,
		unsigned int len, int direction) {

	if (((state->role == CLIENT && direction == RECEIVE)
			|| (state->role == SERVER && direction == SEND))
			&& state->last_message == S_Hello) {
		//@ assert StateAfterServerHello{Pre}(state);
		if (state->kem != DiffieHellman && state->kem != DHE
				&& state->kem != DHEExport && state->kem != ECDH
				&& state->kem != ECDHE && state->kem != CS_RSA
				&& state->kem != RSAExport) {
			//@ assert !CsRequiresServerCertificate{Pre}(state);
#ifdef DEBUG
			fprintf(stderr,
					"The server certificate was rejected because the cipher suite %d did not require it.\n",
					state->kem);
#endif
			return REJECT;
		} else {
			//@ assert sentByServer(state, direction);
			//@ assert CsRequiresServerCertificate{Pre}(state);

			/*@
			 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
			 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));


			 ensures MessageAddedToLog{Here,Pre}(state, msg, len);
			 ensures state->last_message == S_Cert;
			 ensures state->resumption = NO;

			 assigns state->log_length, state->log[\at(state->log_length,Pre)..\at(state->log_length,Pre)+len-1], state->last_message, state->resumption;
			 */
			{
				appendMessageToLog(state, msg, len);

				state->last_message = S_Cert;
				state->resumption = NO;
				//@ assert \at(state->log_length, Pre) + len == \at(state->log_length, Here);
				//@ assert \forall integer i; 0 <= i < len ==> \at(state->log[i + \at(state->log_length,Pre)],Here) == \at(msg[i],Pre);
				//@ assert \forall integer i; 0 <= i < \at(state->log_length, Pre) ==> \at(state->log[i],Here) == \at(state->log[i],Pre);

				//@ assert MessageAddedToLog{Here,Pre}(state, msg, len);
			}
			//@ assert \at(state->last_message,Pre) == S_Hello ;
			//@ assert \at(state->last_message,Here) == S_Cert;
			//@ assert \at(state->resumption,Here) == NO;
			//@ assert HaveSameStateValuesButResumption{Here,Pre}(state);

			//@ assert MessageAddedToLog{Here,Pre}(state,msg, len);
			//@ assert isServerCertificate{Here}(state, msg, len, direction, handshake);
			//@ assert StateAfterServerHello{Pre}(state);

			//@ assert NextStateAfterServerCertificate{Here,Pre}(state, msg, len, direction);

			//@ assert StateAfterServerCertificate{Here}(state);

			return ACCEPT;
		}
	} else {
#ifdef DEBUG
		fprintf(stderr,
				"STATE MONITOR : handle_server_certificate rejected a message of an inappropriate previous state.\n");
#endif
		return REJECT;
	}
}

/*
 * Handles the ServerKeyExchange messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isServerKeyExchange(msg, len, handshake);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success_from_server_certificate :
 assumes
 sentByServer(state, direction) &&
 StateAfterServerCertificate{Pre}(state) &&
 (state->kem == DHE || state->kem && ECDHE || state->kem == DHEExport || state->kem == RSAExport);

 ensures \result == ACCEPT ==> StateAfterServerKeyExchange{Here}(state) ;

 behavior success_from_server_hello :
 assumes
 sentByServer(state, direction) &&
 StateAfterServerHello{Pre}(state) &&
 (state->kem == DHanon || state->kem == ECDHanon);

 ensures \result == ACCEPT ==> StateAfterServerKeyExchange{Here}(state) ;

 behavior failure :
 assumes !(sentByServer(state, direction) &&
 StateAfterServerCertificate{Pre}(state) &&
 (state->kem == DHE || state->kem && ECDHE || state->kem == DHEExport) ) &&
 !( sentByServer(state, direction) &&
 StateAfterServerHello{Pre}(state) &&
 (state->kem == DHanon || state->kem == ECDHanon) );

 ensures \result == REJECT;
 */
int handle_server_key_exchange(STATE_MON *state, unsigned char *msg,
		unsigned int len, int direction) {

	if (((state->role == CLIENT && direction == RECEIVE)
			|| (state->role == SERVER && direction == SEND))
			&& (state->kem != DiffieHellman && state->kem != CS_RSA
					&& state->kem != ECDH)) {

		if (state->last_message != S_Hello && state->last_message != S_Cert) {
			//@ assert !StateAfterServerHello{Pre}(state) && !StateAfterServerCertificate{Pre}(state);
#ifdef DEBUG
			fprintf(stderr,
					"STATE MONITOR : handle_server_key_exchange rejected the SKE because it did not came after the ServerHello or the ServerCertificate.\n");
#endif
			return REJECT;
		} else if (state->last_message == S_Hello && state->kem != DHanon
				&& state->kem != ECDHanon) {
			//@ assert StateAfterServerHello{Pre}(state) && !(state->kem == DHanon || state->kem == ECDHanon);
#ifdef DEBUG
			fprintf(stderr,
					"STATE MONITOR : handle_server_key_exchange rejected a message because it came after a server hello and with a wrong CipherSuite.\n");
#endif
			return REJECT;
		} else if (state->last_message == S_Cert && state->kem != DHE
				&& state->kem != DHEExport && state->kem != ECDHE
				&& state->kem != RSAExport) {
			//@ assert StateAfterServerCertificate{Pre}(state) && !(( state->kem == DHE || state->kem == ECDHE || state->kem == DHEExport || state->kem == RSAExport));
#ifdef DEBUG
			fprintf(stderr,
					"STATE MONITOR : handle_server_key_exchange rejected a message because it came after a ServerCertificate and with a wrong cipher suite.\n");
#endif
			return REJECT;
		} else {

			//@ assert sentByServer(state, direction);
			/*@
			 assert (StateAfterServerHello{Pre}(state) && (state->kem == DHanon || state->kem == ECDHanon))
			 ||
			 (StateAfterServerCertificate{Pre}(state) && ( state->kem == DHE || state->kem == ECDHE || state->kem == DHEExport || state->kem == RSAExport));
			 */

			/*@
			 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));

			 ensures MessageAddedToLog{Here,Pre}(state, msg, len);
			 ensures state->last_message == S_KE;
			 ensures (\at(state->kem,Pre) == DHanon || \at(state->kem,Pre) == ECDHanon) ==> state->resumption == NO;
			 ensures (\at(state->kem,Pre) == DHE || \at(state->kem,Pre) == ECDHE || \at(state->kem,Pre) == DHEExport || \at(state->kem,Pre) == RSAExport) ==> \at(state->resumption,Here) == \at(state->resumption,Pre);

			 assigns state->log_length, state->log[\at(state->log_length,Pre)..\at(state->log_length,Pre)+len-1], state->last_message, state->resumption;
			 */
			{
				appendMessageToLog(state, msg, len);

				state->last_message = S_KE;

				if (state->kem == DHanon || state->kem == ECDHanon) {
					state->resumption = NO;
				}
				//@ assert \at(state->log_length,Here) == \at(state->log_length,Pre) + len ;
				//@ assert (\forall integer i; 0 <= i < \at(state->log_length,Pre) ==> \at(state->log[i],Pre) == \at(state->log[i],Here)) ;
				//@ assert(\forall integer i; 0 <= i < len ==> \at(state->log[i+\at(state->log_length,Pre)],Here) == \at(msg[i],Pre)) ;

				//@ assert MessageAddedToLog{Here,Pre}(state, msg, len);
			}

			//@ assert \at(state->last_message,Here) == S_KE;
			//@ assert HaveSameStateValuesButResumption{Here,Pre}(state);

			//@ assert MessageAddedToLog{Here,Pre}(state,msg, len);
			//@ assert isServerKeyExchange{Here}(msg, len, handshake);
			/*@
			 assert (StateAfterServerHello{Pre}(state) && ( \at(state->kem,Pre) == DHanon || \at(state->kem,Pre) == ECDHanon ) && \at(state->resumption,Here) == NO)
			 ||
			 (StateAfterServerCertificate{Pre}(state) && ( \at(state->kem,Pre) == DHE || \at(state->kem,Pre) == ECDHE || \at(state->kem,Pre) == DHEExport || \at(state->kem,Pre) == RSAExport) && \at(state->resumption,Here) == \at(state->resumption,Pre));
			 */
			//@ assert NextStateAfterServerKeyExchange{Here,Pre}(state, msg, len);
			//@ assert StateAfterServerKeyExchange{Here}(state);
			return ACCEPT;
		}
	} else {
#ifdef DEBUG
		fprintf(stderr,
				"STATE MONITOR : handle_server_key_exchange rejected a message because it came at an inappropriate moment.\n");
#endif
		return REJECT;
	}
}

/*
 * Handles the ServerCertificateRequest messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isServerCertificateRequest(msg, len, handshake);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success_after_server_certificate :
 assumes StateAfterServerCertificate{Pre}(state)
 && sentByServer(state, direction);

 ensures \result == ACCEPT ==> StateAfterServerCertificateRequest{Here}(state) ;

 behavior success_after_server_key_exchange :
 assumes StateAfterServerKeyExchange{Pre}(state)
 && (\at(state->kem,Pre) == DiffieHellman ||
 \at(state->kem,Pre) == ECDH ||
 \at(state->kem,Pre) == DHE ||
 \at(state->kem,Pre) == ECDHE ||
 \at(state->kem,Pre) == RSAExport)
 && sentByServer(state, direction);

 ensures \result == ACCEPT ==> StateAfterServerCertificateRequest{Here}(state) ;

 behavior failure :
 assumes
 !(
 StateAfterServerCertificate{Pre}(state)
 && sentByServer(state, direction)
 ) &&
 !(
 StateAfterServerKeyExchange{Pre}(state)
 && (\at(state->kem,Pre) == DiffieHellman ||
 \at(state->kem,Pre) == ECDH ||
 \at(state->kem,Pre) == DHE ||
 \at(state->kem,Pre) == ECDHE ||
 \at(state->kem,Pre) == RSAExport)
 && sentByServer(state, direction)
 );

 ensures \result == REJECT;
 */
int handle_server_certificate_request(STATE_MON *state, unsigned char *msg,
		unsigned int len, int direction) {

	if (((state->last_message == S_Cert)
			|| (state->last_message == S_KE
					&& (state->kem == DiffieHellman || state->kem == ECDH
							|| state->kem == DHE || state->kem == ECDHE)))
			&& ((state->role == CLIENT && direction == RECEIVE)
					|| (state->role == SERVER && direction == SEND))) {
		/*@
		 assert StateAfterServerCertificate{Pre}(state) ||
		 ( StateAfterServerKeyExchange{Pre}(state) && (\at(state->kem,Pre) == DiffieHellman ||
		 \at(state->kem,Pre) == ECDH ||
		 \at(state->kem,Pre) == DHE ||
		 \at(state->kem,Pre) == ECDHE) ) ;
		 */

		//@ assert sentByServer(state, direction);

		/*@
		 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));

		 ensures MessageAddedToLog{Here,Pre}(state, msg, len);
		 ensures state->last_message == S_CReq;
		 ensures state->client_auth == AUTH_REQUESTED;

		 assigns state->log_length, state->log[\at(state->log_length,Pre)..\at(state->log_length,Pre)+len-1], state->last_message, state->client_auth;
		 */
		{
			appendMessageToLog(state, msg, len);

			state->last_message = S_CReq;
			state->client_auth = AUTH_REQUESTED;
			//@ assert \at(state->log_length,Here) == \at(state->log_length,Pre) + len ;
			//@ assert (\forall integer i; 0 <= i < \at(state->log_length,Pre) ==> \at(state->log[i],Pre) == \at(state->log[i],Here)) ;
			//@ assert(\forall integer i; 0 <= i < len ==> \at(state->log[i+\at(state->log_length,Pre)],Here) == \at(msg[i],Pre)) ;

			//@ assert MessageAddedToLog{Here,Pre}(state, msg, len);
		}

		//@ assert \at(state->last_message,Here) == S_CReq;
		//@ assert HaveSameStateValuesButClientAuth{Here,Pre}(state);
		//@ assert \at(state->client_auth,Here) == AUTH_REQUESTED;

		//@ assert MessageAddedToLog{Here,Pre}(state,msg, len);
		//@ assert isServerCertificateRequest{Here}(msg, len, handshake);
		/*@
		 assert ( StateAfterServerCertificate{Pre}(state) )
		 ||
		 ( StateAfterServerKeyExchange{Pre}(state) && (\at(state->kem,Pre) == DiffieHellman ||
		 \at(state->kem,Pre) == ECDH ||
		 \at(state->kem,Pre) == DHE ||
		 \at(state->kem,Pre) == ECDHE) );
		 */
		//@ assert NextStateAfterServerCertificateRequest{Here,Pre}(state, msg, len);
		//@ assert StateAfterServerCertificateRequest{Here}(state);
		return ACCEPT;

	} else {
#ifdef DEBUG
		fprintf(stderr,
				"STATE MONITOR : Rejected because of a wrong state or a wrong direction of the message.\n");
#endif
		return REJECT;
	}
}

/*
 * Handles the ServerHelloDone messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isServerHelloDone(msg, len, handshake);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success_after_server_certificate_request :
 assumes StateAfterServerCertificate{Pre}(state)
 && (\at(state->kem,Pre) == DiffieHellman ||
 \at(state->kem,Pre) == ECDH ||
 \at(state->kem,Pre) == CS_RSA)
 && sentByServer(state, direction);

 ensures \result == ACCEPT ==> NextStateAfterServerHelloDone{Here,Pre}(state, msg, len) ;

 behavior success_after_server_key_exchange :
 assumes StateAfterServerKeyExchange{Pre}(state)
 && (\at(state->kem,Pre) == DHanon ||
 \at(state->kem,Pre) == ECDHanon ||
 \at(state->kem,Pre) == DHE ||
 \at(state->kem,Pre) == ECDHE ||
 \at(state->kem,Pre) == DHEExport ||
 \at(state->kem,Pre) == RSAExport)
 && sentByServer(state, direction);

 ensures \result == ACCEPT ==> NextStateAfterServerHelloDone{Here,Pre}(state, msg, len) ;

 behavior success_after_server_certificate :
 assumes StateAfterServerCertificateRequest{Pre}(state)
 && (\at(state->kem,Pre) == DiffieHellman ||
 \at(state->kem,Pre) == ECDH ||
 \at(state->kem,Pre) == CS_RSA ||
 \at(state->kem,Pre) == RSAExport ||
 \at(state->kem,Pre) == DHE ||
 \at(state->kem,Pre) == DHEExport ||
 \at(state->kem,Pre) == ECDHE )
 && sentByServer(state, direction);

 ensures \result == ACCEPT ==> NextStateAfterServerHelloDone{Here,Pre}(state, msg, len) ;

 behavior failure :
 assumes
 !(
 StateAfterServerCertificate{Pre}(state)
 && (\at(state->kem,Pre) == DiffieHellman ||
 \at(state->kem,Pre) == ECDH ||
 \at(state->kem,Pre) == CS_RSA)
 && sentByServer(state, direction)
 ) && !(
 StateAfterServerKeyExchange{Pre}(state)
 && (\at(state->kem,Pre) == DHanon ||
 \at(state->kem,Pre) == ECDHanon ||
 \at(state->kem,Pre) == DHE ||
 \at(state->kem,Pre) == ECDHE ||
 \at(state->kem,Pre) == DHEExport ||
 \at(state->kem,Pre) == RSAExport)
 && sentByServer(state, direction)
 ) && !(
 StateAfterServerCertificateRequest{Pre}(state)
 && (\at(state->kem,Pre) == DiffieHellman ||
 \at(state->kem,Pre) == ECDH ||
 \at(state->kem,Pre) == CS_RSA ||
 \at(state->kem,Pre) == RSAExport ||
 \at(state->kem,Pre) == DHE ||
 \at(state->kem,Pre) == DHEExport ||
 \at(state->kem,Pre) == ECDHE )
 && sentByServer(state, direction)
 );

 ensures \result == REJECT;
 */
int handle_server_hello_done(STATE_MON *state, unsigned char *msg,
		unsigned int len, int direction) {

	if ((state->last_message == S_CReq || state->last_message == S_KE
			|| state->last_message == S_Cert)
			&& ((state->role == CLIENT && direction == RECEIVE)
					|| (state->role == SERVER && direction == SEND))) {
		//@ assert sentByServer(state,direction);
		//@ assert ( StateAfterServerCertificate{Pre}(state) || StateAfterServerKeyExchange{Pre}(state) || StateAfterServerCertificateRequest{Pre}(state)) ;
		if (state->last_message == S_Cert && state->kem != DiffieHellman
				&& state->kem != ECDH && state->kem != CS_RSA) {
#ifdef DEBUG
			fprintf(stderr,
					"STATE MONITOR : Rejected because after after a server certificate DH, ECDH or RSA is needed to go to server hello done.\n");
#endif
			return REJECT;
		} else if (state->last_message == S_KE && state->kem != DHanon
				&& state->kem != ECDHanon && state->kem != DHE
				&& state->kem != ECDHE && state->kem != DHEExport
				&& state->kem != RSAExport) {
#ifdef DEBUG
			fprintf(stderr,
					"STATE MONITOR : Rejected because after after a server key exchange DHanon, ECDHanon, DHE, DHEExport, RSAExport or ECDHE is needed to go to server hello done.\n");
#endif
			return REJECT;
		} else if (state->last_message == S_CReq && state->kem != DiffieHellman
				&& state->kem != ECDH && state->kem != CS_RSA
				&& state->kem != DHE && state->kem != ECDHE
				&& state->kem != DHEExport && state->kem != RSAExport) {
#ifdef DEBUG
			fprintf(stderr,
					"STATE MONITOR : Rejected because after after a server certificate request DH, ECDH, RSA, DHE, ECDHE or DHEExport is needed to go to server hello done.\n");
#endif
			return REJECT;
		} else {
			//@ assert sentByServer(state, direction);
			/*@
			 assert ( StateAfterServerCertificate{Pre}(state) && (\at(state->kem,Pre) == DiffieHellman ||
			 \at(state->kem,Pre) == ECDH ||
			 \at(state->kem,Pre) == CS_RSA))
			 ||
			 ( StateAfterServerKeyExchange{Pre}(state) && (\at(state->kem,Pre) == DHanon ||
			 \at(state->kem,Pre) == ECDHanon ||
			 \at(state->kem,Pre) == DHE ||
			 \at(state->kem,Pre) == DHEExport ||
			 \at(state->kem,Pre) == RSAExport ||
			 \at(state->kem,Pre) == ECDHE))
			 ||
			 ( StateAfterServerCertificateRequest{Pre}(state) && (\at(state->kem,Pre) == DiffieHellman ||
			 \at(state->kem,Pre) == ECDH ||
			 \at(state->kem,Pre) == CS_RSA ||
			 \at(state->kem,Pre) == DHE ||
			 \at(state->kem,Pre) == DHEExport ||
			 \at(state->kem,Pre) == RSAExport ||
			 \at(state->kem,Pre) == ECDHE ));
			 */

			/*@
			 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));

			 ensures MessageAddedToLog{Here,Pre}(state, msg, len);
			 ensures state->last_message == S_HD;
			 ensures (\at(state->last_message,Pre) == S_Cert || \at(state->last_message,Pre) == S_KE) ==> state->client_auth == NO_AUTH;
			 ensures (\at(state->last_message,Pre) == S_CReq) ==> ( state->client_auth == \at(state->client_auth,Pre) );
			 assigns state->log_length, state->log[\at(state->log_length,Pre)..\at(state->log_length,Pre)+len-1], state->last_message, state->client_auth;
			 */
			{
				appendMessageToLog(state, msg, len);

				if (state->last_message == S_Cert
						|| state->last_message == S_KE)
					state->client_auth = NO_AUTH;

				state->last_message = S_HD;
				//@ assert \at(state->log_length,Here) == \at(state->log_length,Pre) + len ;
				//@ assert (\forall integer i; 0 <= i < \at(state->log_length,Pre) ==> \at(state->log[i],Pre) == \at(state->log[i],Here)) ;
				//@ assert(\forall integer i; 0 <= i < len ==> \at(state->log[i+\at(state->log_length,Pre)],Here) == \at(msg[i],Pre)) ;

				//@ assert MessageAddedToLog{Here,Pre}(state, msg, len);

			}
			//@ assert \at(state->last_message,Here) == S_HD;
			//@ assert HaveSameStateValuesButClientAuth{Here,Pre}(state);

			//@ assert MessageAddedToLog{Here,Pre}(state,msg, len);
			//@ assert isServerHelloDone{Here}(msg, len, handshake);

			//@ assert StateAfterServerCertificate{Pre}(state) ==> \at(state->client_auth,Here) == NO_AUTH;
			//@ assert StateAfterServerKeyExchange{Pre}(state) ==> \at(state->client_auth,Here) == NO_AUTH;
			//@ assert StateAfterServerCertificateRequest{Pre}(state) ==> \at(state->client_auth,Here) == \at(state->client_auth,Pre);

			/*@
			 assert ( StateAfterServerCertificate{Pre}(state) && (\at(state->kem,Pre) == DiffieHellman ||
			 \at(state->kem,Pre) == ECDH ||
			 \at(state->kem,Pre) == CS_RSA)
			 && \at(state->client_auth,Here) == NO_AUTH)
			 ||
			 ( StateAfterServerKeyExchange{Pre}(state) && (\at(state->kem,Pre) == DHanon ||
			 \at(state->kem,Pre) == ECDHanon ||
			 \at(state->kem,Pre) == DHE ||
			 \at(state->kem,Pre) == DHEExport ||
			 \at(state->kem,Pre) == RSAExport ||
			 \at(state->kem,Pre) == ECDHE)
			 && \at(state->client_auth,Here) == NO_AUTH)
			 ||
			 ( StateAfterServerCertificateRequest{Pre}(state) && (\at(state->kem,Pre) == DiffieHellman ||
			 \at(state->kem,Pre) == ECDH ||
			 \at(state->kem,Pre) == CS_RSA ||
			 \at(state->kem,Pre) == DHE ||
			 \at(state->kem,Pre) == DHEExport ||
			 \at(state->kem,Pre) == RSAExport ||
			 \at(state->kem,Pre) == ECDHE )
			 && \at(state->client_auth,Here) == \at(state->client_auth,Pre));
			 */
			//@ assert NextStateAfterServerHelloDone{Here,Pre}(state, msg, len);
			//@ assert StateAfterServerHelloDone{Here}(state);
			return ACCEPT;
		}
	} else {
#ifdef DEBUG
		fprintf(stderr,
				"STATE MONITOR : Rejected because of a mismatch with the previous state.\n");
#endif
		return REJECT;
	}
}

/*
 * Handles the ClientCertificates messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isClientCertificate(state, msg, len, direction, handshake);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success :
 assumes  StateAfterServerHelloDone{Pre}(state)
 && state->client_auth == AUTH_REQUESTED
 && sentByClient(state, direction);

 ensures \result == ACCEPT ==> NextStateAfterClientCertificate{Here,Pre}(state, msg, len, direction) ;

 behavior failure :
 assumes
 !(	StateAfterServerHelloDone{Pre}(state)
 && state->client_auth == AUTH_REQUESTED
 && sentByClient(state, direction) );

 ensures \result == REJECT;
 */
int handle_client_certificate(STATE_MON *state, unsigned char *msg,
		unsigned int len, int direction) {

	if (state->last_message == S_HD && state->client_auth == AUTH_REQUESTED
			&& ((state->role == CLIENT && direction == SEND)
					|| (state->role == SERVER && direction == RECEIVE))) {

		//@ assert sentByClient(state, direction);
		//@ assert StateAfterServerHelloDone{Pre}(state);
		//@ assert \at(state->client_auth,Pre) == AUTH_REQUESTED;

		/*@
		 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
		 requires isClientCertificate(state, msg, len, direction, handshake);

		 ensures MessageAddedToLog{Here,Pre}(state, msg, len);
		 ensures state->last_message == C_Cert;
		 ensures (isEmptyClientCertificate{Pre}(state, msg, len, direction) && \at(state->client_auth,Here) == AUTH_NOT_PROVIDED )
		 ||
		 (isNonEmptyClientCertificate{Pre}(state, msg, len, direction) && \at(state->client_auth,Here) == AUTH_PROVIDED );

		 assigns state->log_length, state->log[\at(state->log_length,Pre)..\at(state->log_length,Pre)+len-1], state->client_auth, state->last_message;
		 */
		{
			appendMessageToLog(state, msg, len);
			state->client_auth = checkClientCertificate(state, msg, len,
					direction);

			state->last_message = C_Cert;

			//@ assert \at(state->log_length,Pre) + len == \at(state->log_length,Here);
			//@ assert \forall integer i; 0 <= i < \at(state->log_length, Pre) ==> \at(state->log[i],Pre) == \at(state->log[i],Here);
			//@ assert \forall integer i; 0 <= i < len ==> \at(state->log[i + \at(state->log_length,Pre)],Here) == \at(msg[i],Pre);

			//@ assert MessageAddedToLog{Here,Pre}(state, msg, len);
		}
		//@ assert \at(state->last_message,Here) == C_Cert;
		//@ assert HaveSameStateValuesButClientAuth{Here,Pre}(state);

		//@ assert MessageAddedToLog{Here,Pre}(state,msg, len);
		//@ assert isClientCertificate{Here}(state, msg, len, direction, handshake);
		//@ assert StateAfterServerHelloDone{Pre}(state);
		//@ assert \at(state->client_auth,Pre) == AUTH_REQUESTED;

		//@ assert NextStateAfterClientCertificate{Here,Pre}(state, msg, len, direction);
		//@ assert StateAfterClientCertificate{Here}(state);
		return ACCEPT;

	} else {
#ifdef DEBUG
		fprintf(stderr,
				"STATE MONITOR : Rejected because of a mismatch with the previous state.\n");
		fprintf(stderr, "Last msg : %d, while expected : %d\n",
				state->last_message, S_HD);
		fprintf(stderr, "Client authentication : %d, while expected : %d\n",
				state->client_auth, AUTH_REQUESTED);
#endif
		return REJECT;
	}
}

/*
 * Handles the ClientKeyExchange messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isClientKeyExchange(msg, len, handshake);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success_after_server_hello_done :
 assumes  StateAfterServerHelloDone{Pre}(state)
 && state->client_auth == NO_AUTH
 && sentByClient(state, direction);

 ensures \result == ACCEPT ==> NextStateAfterClientKeyExchange{Here,Pre}(state, msg, len) ;

 behavior success_after_client_certificate :
 assumes StateAfterClientCertificate{Pre}(state) &&
 (state->client_auth == AUTH_PROVIDED || state->client_auth == AUTH_NOT_PROVIDED)
 && sentByClient(state, direction);

 ensures \result == ACCEPT ==> NextStateAfterClientKeyExchange{Here,Pre}(state, msg, len) ;

 behavior failure :
 assumes
 !(
 StateAfterServerHelloDone{Pre}(state)
 && state->client_auth == NO_AUTH
 && sentByClient(state, direction)
 ) &&
 !(
 StateAfterClientCertificate{Pre}(state) &&
 (state->client_auth == AUTH_PROVIDED || state->client_auth == AUTH_NOT_PROVIDED)
 && sentByClient(state, direction)
 );

 ensures \result == REJECT;
 */
int handle_client_key_exchange(STATE_MON *state, unsigned char *msg,
		unsigned int len, int direction) {

	if (((state->last_message == S_HD && state->client_auth == NO_AUTH)
			|| (state->last_message == C_Cert
					&& (state->client_auth == AUTH_PROVIDED
							|| state->client_auth == AUTH_NOT_PROVIDED)))
			&& ((state->role == CLIENT && direction == SEND)
					|| (state->role == SERVER && direction == RECEIVE))) {

		/*@
		 assert (StateAfterClientCertificate{Pre}(state) && (\at(state->client_auth,Pre) == AUTH_PROVIDED || \at(state->client_auth,Pre) == AUTH_NOT_PROVIDED))
		 ||
		 (StateAfterServerHelloDone{Pre}(state) && ( \at(state->client_auth,Pre) == NO_AUTH ));
		 */
		//@ assert sentByClient(state, direction);
		/*@
		 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));

		 ensures MessageAddedToLog{Here,Pre}(state, msg, len);
		 ensures state->last_message == C_KE;
		 assigns state->log_length, state->log[\at(state->log_length,Pre)..\at(state->log_length,Pre)+len-1], state->last_message;
		 */
		{
			appendMessageToLog(state, msg, len);
			state->last_message = C_KE;
			//@ assert \at(state->log_length,Here) == \at(state->log_length,Pre) + len ;
			//@ assert (\forall integer i; 0 <= i < \at(state->log_length,Pre) ==> \at(state->log[i],Pre) == \at(state->log[i],Here)) ;
			//@ assert(\forall integer i; 0 <= i < len ==> \at(state->log[i+\at(state->log_length,Pre)],Here) == \at(msg[i],Here)) ;

			//@ assert MessageAddedToLog{Here,Pre}(state, msg, len);
		}
		//@ assert \at(state->last_message,Here) == C_KE;
		//@ assert HaveSameStateValues{Here,Pre}(state);

		//@ assert MessageAddedToLog{Here,Pre}(state,msg, len);
		//@ assert isClientKeyExchange{Here}(msg, len, handshake);
		/*@
		 assert (StateAfterClientCertificate{Pre}(state) && (\at(state->client_auth,Pre) == AUTH_PROVIDED || \at(state->client_auth,Pre) == AUTH_NOT_PROVIDED))
		 ||
		 (StateAfterServerHelloDone{Pre}(state) && ( \at(state->client_auth,Pre) == NO_AUTH ));
		 */

		//@ assert NextStateAfterClientKeyExchange{Here,Pre}(state, msg, len);
		//@ assert StateAfterClientKeyExchange{Here}(state);
		return ACCEPT;

	} else {
#ifdef DEBUG
		fprintf(stderr,
				"STATE MONITOR : Rejected because the state does not match what expected.\n");
#endif
		return REJECT;
	}
}

/*
 * Handles the ClientCertificateVerify messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isClientCertificateVerify(msg, len, handshake);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success :
 assumes  StateAfterClientKeyExchange{Pre}(state)
 && sentByClient(state, direction)
 && \at(state->client_auth,Pre) == AUTH_PROVIDED;

 ensures \result == ACCEPT ==> StateAfterClientCertificateVerify{Here}(state);

 behavior failure :
 assumes !( StateAfterClientKeyExchange{Pre}(state)
 && sentByClient(state, direction)
 && \at(state->client_auth,Pre) == AUTH_PROVIDED);

 ensures \result == REJECT;
 */
int handle_client_certificate_verify(STATE_MON *state, unsigned char *msg,
		unsigned int len, int direction) {

	if (state->last_message == C_KE && state->client_auth == AUTH_PROVIDED
			&& ((state->role == CLIENT && direction == SEND)
					|| (state->role == SERVER && direction == RECEIVE))) {
		//@ assert sentByClient(state, direction);
		//@ assert StateAfterClientKeyExchange{Pre}(state);
		//@ assert \at(state->client_auth,Pre) == AUTH_PROVIDED;

		/*@
		 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));

		 ensures MessageAddedToLog{Here,Pre}(state, msg, len);
		 ensures state->last_message == C_CV;
		 assigns state->log_length, state->log[\at(state->log_length,Pre)..\at(state->log_length,Pre)+len-1], state->last_message;
		 */
		{
			appendMessageToLog(state, msg, len);

			state->last_message = C_CV;

			//@ assert \at(state->log_length,Pre) + len == \at(state->log_length,Here);
			//@ assert \forall integer i; 0 <= i < \at(state->log_length, Pre) ==> \at(state->log[i],Pre) == \at(state->log[i],Here);
			//@ assert \forall integer i; 0 <= i < len ==> \at(state->log[i + \at(state->log_length,Pre)],Here) == \at(msg[i],Pre);

			//@ assert MessageAddedToLog{Here,Pre}(state, msg, len);
		}
		//@ assert \at(state->last_message,Here) == C_CV;
		//@ assert HaveSameStateValues{Here,Pre}(state);

		// @ assert \at(state->log_length,Here) == \at(state->log_length,Pre) + len ;
		// @ assert (\forall integer i; 0 <= i < \at(state->log_length,Pre) ==> \at(state->log[i],Pre) == \at(state->log[i],Here)) ;
		// @ assert(\forall integer i; 0 <= i < len ==> \at(state->log[i+\at(state->log_length,Pre)],Here) == \at(msg[i],Here)) ;

		//@ assert MessageAddedToLog{Here,Pre}(state,msg, len);
		//@ assert isClientCertificateVerify{Here}(msg, len, handshake);
		//@ assert StateAfterClientKeyExchange{Pre}(state);
		//@ assert \at(state->client_auth,Pre) == AUTH_PROVIDED;

		//@ assert NextStateAfterClientCertificateVerify{Here,Pre}(state, msg, len);
		//@ assert StateAfterClientCertificateVerify{Here}(state);

		return ACCEPT;
	} else {
#ifdef DEBUG
		fprintf(stderr,
				"STATE MONITOR : Rejected because the state does not match what expected.\n");
#endif
		return REJECT;

	}
}

/*
 * Handles the ServerNewSessionTicket messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isServerNewSessionTicket(msg, len, handshake);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success_after_server_hello :
 assumes  StateAfterServerHello{Pre}(state)
 && sentByServer(state, direction);

 ensures \result == ACCEPT ==> StateAfterServerNewSessionTicket{Here}(state);

 behavior success_after_client_fin :
 assumes  StateAfterClientFin{Pre}(state)
 && sentByServer(state, direction);

 ensures \result == ACCEPT ==> StateAfterServerNewSessionTicket{Here}(state);

 behavior failure :
 assumes !( StateAfterServerHello{Pre}(state)
 && sentByServer(state, direction)
 ) &&
 !(
 StateAfterClientFin{Pre}(state)
 && sentByServer(state, direction)
 );

 ensures \result == REJECT;
 */
int handle_server_new_session_ticket(STATE_MON *state, unsigned char *msg,
		unsigned int len, int direction) {

	if ((state->last_message == S_Hello
			|| (state->last_message == C_Fin && state->resumption == NO))
			&& ((state->role == SERVER && direction == SEND)
					|| (state->role == CLIENT && direction == RECEIVE))) {
		//@ assert sentByServer(state, direction);
		//@ assert ( StateAfterServerHello{Pre}(state)  ) || (StateAfterClientFin{Pre}(state) && \at(state->resumption,Pre) == NO);

		/*@
		 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));

		 ensures MessageAddedToLog{Here,Pre}(state, msg, len);
		 ensures state->last_message == ServerNewSessionTicket;
		 ensures \at(state->last_message,Pre) == S_Hello ==> (state->resumption == YES);
		 ensures \at(state->last_message,Pre) == C_Fin ==> (state->resumption,Pre) == NO;
		 ensures state->ntick == YES;
		 assigns state->log_length, state->log[\at(state->log_length,Pre)..\at(state->log_length,Pre)+len-1], state->last_message, state->resumption, state->ntick;
		 */
		{
			appendMessageToLog(state, msg, len);
			if (state->last_message == S_Hello)
				state->resumption = YES;

			state->last_message = ServerNewSessionTicket;
			state->ntick = YES;
			//@ assert \at(state->log_length,Pre) + len == \at(state->log_length,Here);
			//@ assert \forall integer i; 0 <= i < \at(state->log_length, Pre) ==> \at(state->log[i],Pre) == \at(state->log[i],Here);
			//@ assert \forall integer i; 0 <= i < len ==> \at(state->log[i + \at(state->log_length,Pre)],Here) == \at(msg[i],Pre);

			//@ assert MessageAddedToLog{Here,Pre}(state, msg, len);
		}
		//@ assert \at(state->last_message,Here) == ServerNewSessionTicket;
		//@ assert HaveSameStateValuesButResumptionAndTicket{Here,Pre}(state);

		//@ assert MessageAddedToLog{Here,Pre}(state,msg, len);
		//@ assert isServerNewSessionTicket{Here}(msg, len, handshake);
		//@ assert \at(state->ntick,Here) == YES;
		//@ assert ( StateAfterServerHello{Pre}(state) && \at(state->resumption,Here) == YES ) || (StateAfterClientFin{Pre}(state) && \at(state->resumption,Pre) == \at(state->resumption,Here));

		//@ assert NextStateAfterServerNewSessionTicket{Here,Pre}(state, msg, len);
		//@ assert StateAfterServerNewSessionTicket{Here}(state);
		return ACCEPT;
	} else {
#ifdef DEBUG
		fprintf(stderr,
				"STATE MONITOR : Rejected because the state does not match what expected.\n");
#endif
		return REJECT;
	}
}

/*
 * Handles the ServerCCS messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isServerChangeCipherSpecs(state, msg, len, direction, change_cipher_spec);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success_after_resumption_no_session_ticket :
 assumes StateAfterServerHello{Pre}(state)
 && sentByServer(state, direction);

 ensures \result == ACCEPT ==> StateAfterServerCCS{Here}(state);

 behavior success_after_resumption_and_session_ticket :
 assumes StateAfterServerNewSessionTicket{Pre}(state)
 && sentByServer(state, direction);

 ensures \result == ACCEPT ==> StateAfterServerCCS{Here}(state);

 behavior success_no_resumption :
 assumes StateAfterClientFin{Pre}(state)
 && \at(state->resumption,Pre) == NO
 && sentByServer(state, direction);

 ensures \result == ACCEPT ==> StateAfterServerCCS{Here}(state) ;

 behavior failure :
 assumes
 !(
 StateAfterServerHello{Pre}(state)
 && sentByServer(state, direction)
 ) &&
 !(
 StateAfterServerNewSessionTicket{Pre}(state)
 && sentByServer(state, direction)
 )&&
 !(
 StateAfterClientFin{Pre}(state)
 && \at(state->resumption,Pre) == NO
 && sentByServer(state, direction)
 );

 ensures \result == REJECT;
 */
int handle_server_change_cipher_specs(STATE_MON *state, unsigned char *msg,
		unsigned int len, int direction) {

	if ((state->last_message == S_Hello
			|| (state->last_message == C_Fin && state->resumption == NO)
			|| state->last_message == ServerNewSessionTicket)
			&& ((state->role == CLIENT && direction == RECEIVE)
					|| (state->role == SERVER && direction == SEND))) {
		// @ assert messageFitsInLog(state, len);
		// @ assert sentByServer(state, direction);
		/*@
		 assert (StateAfterServerHello{Pre}(state) )
		 ||
		 (StateAfterClientFin{Pre}(state) && \at(state->resumption,Pre) == NO )
		 ||
		 (StateAfterServerNewSessionTicket{Pre}(state) );
		 */

		/*@
		 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));

		 ensures state->last_message == S_CCS;
		 ensures (\at(state->last_message,Pre) == S_Hello) ==> state->resumption == YES && state->ntick == NO;
		 ensures (\at(state->last_message,Pre) == C_Fin) ==> state->ntick == NO && state->resumption == NO;
		 ensures (\at(state->last_message,Pre) == ServerNewSessionTicket) ==> (\at(state->resumption,Pre) == state->resumption && \at(state->ntick,Pre) == state->ntick);

		 assigns state->last_message, state->resumption, state->ntick;
		 */
		{

			if (state->last_message == S_Hello) {
				state->resumption = YES;
				state->ntick = NO;
			} else if (state->last_message == C_Fin) {
				state->ntick = NO;
			}

			state->last_message = S_CCS;

		}

		//@ assert \at(state->last_message,Here) == S_CCS;
		//@ assert HaveSameStateValuesButResumptionAndTicket{Here,Pre}(state);

		//@ assert isServerChangeCipherSpecs{Here}(state, msg, len, direction, change_cipher_spec);
		/*@
		 assert (StateAfterServerHello{Pre}(state) && \at(state->resumption,Here) == YES && \at(state->ntick,Here) == NO )
		 ||
		 (StateAfterClientFin{Pre}(state) && \at(state->resumption,Pre) == NO && \at(state->resumption,Pre) == \at(state->resumption,Here)  && \at(state->ntick,Here) == NO )
		 ||
		 (StateAfterServerNewSessionTicket{Pre}(state) && \at(state->resumption,Here) == \at(state->resumption,Pre) && \at(state->ntick,Here) == \at(state->ntick,Pre));
		 */

		//@ assert NextStateAfterServerCCS{Here,Pre}(state, msg, len, direction);
		//@ assert StateAfterServerCCS{Here}(state);
		return ACCEPT;
	} else {
#ifdef DEBUG
		fprintf(stderr,
				"STATE MONITOR : Rejected because the state does not match what expected.\n");
#endif
		return REJECT;
	}
}

/*
 * Handles the ServerFinished messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isServerFin(state, msg, len, direction, handshake);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success :
 assumes StateAfterServerCCS{Pre}(state)
 && sentByServer(state, direction);

 ensures \result == ACCEPT ==> StateAfterServerFin{Here}(state) ;

 behavior failure :
 assumes !( StateAfterServerCCS{Pre}(state)
 && sentByServer(state, direction) );

 ensures \result == REJECT;
 */
int handle_server_fin(STATE_MON *state, unsigned char *msg, unsigned int len,
		int direction) {

	if (state->last_message == S_CCS
			&& ((state->role == CLIENT && direction == RECEIVE)
					|| (state->role == SERVER && direction == SEND))) {
		// @ assert sentByServer(state, direction);
		// @ assert StateAfterServerCCS{Pre}(state) ;

		/*@
		 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));

		 ensures MessageAddedToLog{Here,Pre}(state, msg, len);
		 ensures state->last_message == S_Fin;

		 assigns state->log_length, state->log[\at(state->log_length,Pre)..\at(state->log_length,Pre)+len-1], state->last_message;
		 */
		{
			appendMessageToLog(state, msg, len);
			state->last_message = S_Fin;

			//@ assert \at(state->log_length,Pre) + len == \at(state->log_length,Here);
			//@ assert \forall integer i; 0 <= i < \at(state->log_length, Pre) ==> \at(state->log[i],Pre) == \at(state->log[i],Here);
			//@ assert \forall integer i; 0 <= i < len ==> \at(state->log[i + \at(state->log_length,Pre)],Here) == \at(msg[i],Pre);

			//@ assert MessageAddedToLog{Here,Pre}(state, msg, len);
		}

		//@ assert \at(state->last_message,Here) == S_Fin;
		//@ assert HaveSameStateValues{Here,Pre}(state);

		//@ assert MessageAddedToLog{Here,Pre}(state,msg, len);
		//@ assert isServerFin{Here}(state, msg, len, direction, handshake);
		//@ assert StateAfterServerCCS{Pre}(state);

		//@ assert NextStateAfterServerFin{Here,Pre}(state, msg, len, direction);
		//@ assert StateAfterServerFin{Here}(state);

		return ACCEPT;
	} else {
#ifdef DEBUG
		fprintf(stderr,
				"STATE MONITOR : Rejected because the state does not match what expected.\n");
#endif
		return REJECT;
	}
}

/*
 * Handles the ClientCCS messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isClientChangeCipherSpecs(state, msg, len, direction, change_cipher_spec);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success_after_resumption :
 assumes StateAfterServerFin(state)
 && state->resumption == YES
 && sentByClient(state, direction);

 ensures \result == ACCEPT ==> StateAfterClientCCSLastMsg{Here}(state) ;

 behavior success_after_client_certificate_verify :
 assumes StateAfterClientCertificateVerify{Pre}(state)
 && sentByClient(state, direction);

 ensures \result == ACCEPT ==> StateAfterClientCCS{Here}(state) ;

 behavior success_after_client_key_exchange :
 assumes StateAfterClientKeyExchange{Pre}(state)
 && (\at(state->client_auth,Pre) == NO_AUTH || \at(state->client_auth,Pre) == AUTH_NOT_PROVIDED )
 && sentByClient(state, direction);

 ensures \result == ACCEPT ==> StateAfterClientCCS{Here}(state) ;

 behavior success_after_client_certificate :
 assumes StateAfterClientCertificate{Pre}(state)
 && (\at(state->kem,Pre) == DiffieHellman || \at(state->kem,Pre) == ECDH )
 && sentByClient(state, direction);

 ensures \result == ACCEPT ==> StateAfterClientCCS{Here}(state) ;

 behavior failure :
 assumes
 !sentByClient(state, direction)	||
 (
 !(
 StateAfterClientCertificate{Pre}(state)
 && (\at(state->kem,Pre) == DiffieHellman || \at(state->kem,Pre) == ECDH )
 )
 &&
 !(
 StateAfterClientKeyExchange{Pre}(state)
 && (\at(state->client_auth,Pre) == NO_AUTH || \at(state->client_auth,Pre) == AUTH_NOT_PROVIDED )
 )
 &&
 !(
 StateAfterClientCertificateVerify{Pre}(state)
 )
 &&
 !(
 StateAfterServerFin(state)
 && state->resumption == YES
 )
 );

 ensures \result == REJECT;
 */
int handle_client_change_cipher_specs(STATE_MON *state, unsigned char *msg,
		unsigned int len, int direction) {

	if ((state->last_message == C_CV
			|| (state->last_message == C_KE
					&& (state->client_auth == AUTH_NOT_PROVIDED
							|| state->client_auth == NO_AUTH))
			|| (state->last_message == C_Cert
					&& (state->kem == DiffieHellman || state->kem == ECDH)))
			&& ((state->role == SERVER && direction == RECEIVE)
					|| (state->role == CLIENT && direction == SEND))) {
		//@ assert sentByClient(state, direction);

		/*@
		 assert (StateAfterClientCertificateVerify{Pre}(state)  )
		 ||
		 (StateAfterClientKeyExchange{Pre}(state) && (\at(state->client_auth,Pre) == NO_AUTH || \at(state->client_auth,Pre) == AUTH_NOT_PROVIDED ))
		 ||
		 (StateAfterClientCertificate{Pre}(state) && (\at(state->kem,Pre) == DiffieHellman || \at(state->kem,Pre) == ECDH) );
		 */

		/*@
		 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));

		 ensures state->last_message == C_CCS;
		 ensures state->resumption == NO;
		 assigns state->resumption, state->last_message;
		 */
		{

			state->last_message = C_CCS;
			state->resumption = NO;

		}
		//@ assert \at(state->last_message,Here) == C_CCS;
		//@ assert HaveSameStateValuesButResumption{Here,Pre}(state);
		//@ assert \at(state->resumption,Here) == NO;
		//@ assert isClientChangeCipherSpecs{Here}(state, msg, len, direction, change_cipher_spec);
		/*@
		 assert (StateAfterClientCertificateVerify{Pre}(state)  )
		 ||
		 (StateAfterClientKeyExchange{Pre}(state) && (\at(state->client_auth,Pre) == NO_AUTH || \at(state->client_auth,Pre) == AUTH_NOT_PROVIDED ))
		 ||
		 (StateAfterClientCertificate{Pre}(state) && (\at(state->kem,Pre) == DiffieHellman || \at(state->kem,Pre) == ECDH) );
		 */

		//@ assert NextStateAfterClientCCS{Here,Pre}(state, msg, len, direction);
		//@ assert StateAfterClientCCS{Here}(state);
		return ACCEPT;
	} else if (state->last_message == S_Fin && state->resumption == YES
			&& ((state->role == SERVER && direction == RECEIVE)
					|| (state->role == CLIENT && direction == SEND))) {
		//@ assert sentByClient(state, direction);
		//@ assert StateAfterServerFin(state) && state->resumption == YES;

		/*@
		 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));

		 ensures state->last_message == C_CCS;
		 assigns state->last_message;
		 */
		{
			state->last_message = C_CCS;

		}

		//@ assert \at(state->last_message,Here) == C_CCS								;
		//@ assert HaveSameStateValues{Here,Pre}(state)									;
		//@ assert isClientChangeCipherSpecs{Here}(state, msg, len, direction, change_cipher_spec);
		//@ assert (StateAfterServerFin{Pre}(state) && \at(state->resumption,Here) == YES );
		//@ assert \at(state->resumption,Pre) == YES;
		//@ assert NextStateAfterClientCCSLastMsg{Here,Pre}(state,msg, len, direction);
		//@ assert StateAfterClientCCSLastMsg{Here}(state);
		return ACCEPT;
	} else {
#ifdef DEBUG
		fprintf(stderr,
				"STATE MONITOR : Rejected because the state does not match what expected.\n");
#endif
		return REJECT;
	}
}

/*
 * Handles the ClientFinished messages received or sent and updates the state monitor accordingly.
 * Returns REJECT if something wrong was detected, ACCEPT otherwise.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isClientFin(state, msg, len, direction, handshake);
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;

 behavior success_last_message :
 assumes StateAfterClientCCS{Pre}(state)
 && \at(state->resumption,Pre) == YES
 && sentByClient(state, direction);

 ensures \result == ACCEPT ==> StateAfterClientFinLastMsg{Here}(state) ;

 behavior success_not_last_message :
 assumes StateAfterClientCCS{Pre}(state)
 && \at(state->resumption,Pre) == NO
 && sentByClient(state, direction);

 ensures \result == ACCEPT ==> StateAfterClientFin{Here}(state) ;

 behavior failure :
 assumes !( StateAfterClientCCS{Pre}(state)
 && sentByClient(state, direction) );

 ensures \result == REJECT;
 */
int handle_client_fin(STATE_MON *state, unsigned char *msg, unsigned int len,
		int direction) {

	if (state->last_message == C_CCS
			&& ((state->role == CLIENT && direction == SEND)
					|| (state->role == SERVER && direction == RECEIVE))) {
		//@ assert sentByClient(state, direction);
		//@ assert StateAfterClientCCS{Pre}(state) || StateAfterClientCCSLastMsg{Pre}(state);

		appendMessageToLog(state, msg, len);

		state->last_message = C_Fin;

		//@ assert \at(state->last_message,Here) == C_Fin;
		//@ assert HaveSameStateValues{Here,Pre}(state);

		//@ assert \at(state->log_length,Here) == \at(state->log_length,Pre) + len ;
		//@ assert (\forall integer i; 0 <= i < \at(state->log_length,Pre) ==> \at(state->log[i],Pre) == \at(state->log[i],Here)) ;
		//@ assert(\forall integer i; 0 <= i < len ==> \at(state->log[i+\at(state->log_length,Pre)],Here) == \at(msg[i],Here)) ;

		// Takes about 2min to Alt-ergo to prove MessageAddedToLog..., z3 still not done after 5min
		//@ assert MessageAddedToLog{Here,Pre}(state, msg, len);
		//@ assert isClientFin{Here}(state, msg, len, direction, handshake);
		//@ assert ( StateAfterClientCCS{Pre}(state) || StateAfterClientCCSLastMsg{Pre}(state) );

		/*@ assert ( NextStateAfterClientFinLastMsg{Here,Pre}(state, msg, len, direction) )
		 ||
		 ( NextStateAfterClientFin{Here,Pre}(state, msg, len, direction) );
		 */
		//@ assert StateAfterClientFin{Here}(state) || StateAfterClientFinLastMsg{Here}(state);
		return ACCEPT;
	} else {
#ifdef DEBUG
		fprintf(stderr,
				"STATE MONITOR : Rejected because the state does not match what expected.\n");
#endif
		return REJECT;
	}
}

/*@
 requires \valid(state);

 ensures StateAfterInitialState{Here}(state);
 */
void ssl3_initialize_state_mon(STATE_MON *state) {
	fprintf(stderr, "STATE MONITOR : initializing \n");
	state->role = UNDEFINED_ROLE;
	state->version = UNDEFINED_PV;
	state->kem = UNDEFINED_KEM;
	state->client_auth = UNDEFINED_AUTH;
	state->resumption = UNDEFINED;
	state->last_message = UNDEFINED_TYPE;
	state->log = malloc(sizeof(char) * 10000);
	if (!state->log) {
		fprintf(stderr, "Error initializing state monitor log\n");
		abort();
	}
	state->log_length = 0;
	state->renegotiation = UNDEFINED;
	state->ntick = UNDEFINED;
	state->state_mon_initialized = YES;
	state->max_log_size = 10000;
	fprintf(stderr, "STATE MONITOR : initializing \n");
}

void ssl3_realloc_state_mon_log(STATE_MON *state) {
	unsigned char * new_log_ptr = realloc(state->log, 2 * state->max_log_size);
	state->max_log_size = state->max_log_size * 2;
	if (new_log_ptr == NULL) {
		fprintf(stderr,
				"STATE MONITOR ERROR : could not reallocate enough space for the log, aborting.\n");
	} else {
		state->log = new_log_ptr;
	}
}

/*
 * Main state monitor function.
 */
/*@
 requires \valid(state) && \valid(msg+(0..(len-1))) && \valid(state->log+(0..state->log_length+len-1));
 requires \separated(msg+(0..(len-1)), state+(0..(sizeof(state)-1)), state->log+(0..state->log_length+len-1));
 requires isValidState{Pre}(state);

 ensures \result == ACCEPT || \result == REJECT;
 ensures ( \result == ACCEPT ==> isValidState(state) );
 */
int ssl3_next_message(STATE_MON *state, unsigned char* msg, unsigned int len,
		int direction, unsigned char content_type) {
	int res;

	switch (getMessageType(state, msg, len, direction, content_type)) {
	case C_Hello:
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : Got ClientHello.\n");
#endif
		res = handle_client_hello(state, msg, len, direction);
		//@ assert ( res == ACCEPT || res == REJECT );
		//@ assert ( res == ACCEPT ) ==> StateAfterClientHello(state);
		//@ assert StateAfterClientHello(state) ==> isValidState(state);
		//@ assert ( res == ACCEPT && isValidState(state) ) || (res == REJECT ) ;
		return res;
	case S_Hello:
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : Got ServerHello.\n");
#endif

		res = handle_server_hello(state, msg, len, direction);
		//@ assert ( res == ACCEPT || res == REJECT );
		//@ assert ( res == ACCEPT ) ==> StateAfterServerHello(state);
		//@ assert StateAfterServerHello(state) ==> isValidState(state);
		//@ assert ( res == ACCEPT && isValidState{Here}(state) ) || (res == REJECT ) ;
		return res;
	case S_Cert:
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : Got ServerCertificate.\n");
#endif
		res = handle_server_certificate(state, msg, len, direction);
		//@ assert ( res == ACCEPT || res == REJECT );
		//@ assert ( res == ACCEPT ) ==> StateAfterServerCertificate(state);
		//@ assert StateAfterServerCertificate(state) ==> isValidState(state);
		//@ assert ( res == ACCEPT && isValidState(state) ) || (res == REJECT ) ;
		return res;
	case S_KE:
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : Got ServerKeyExchange.\n");
#endif
		res = handle_server_key_exchange(state, msg, len, direction);
		//@ assert ( res == ACCEPT || res == REJECT );
		//@ assert ( res == ACCEPT ) ==> StateAfterServerKeyExchange(state);
		//@ assert StateAfterServerKeyExchange(state) ==> isValidState(state);
		//@ assert ( res == ACCEPT && isValidState(state) ) || (res == REJECT ) ;
		return res;
	case S_CReq:
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : Got ServerCertificateRequest.\n");
#endif
		res = handle_server_certificate_request(state, msg, len, direction);
		//@ assert ( res == ACCEPT || res == REJECT );
		//@ assert ( res == ACCEPT ) ==> StateAfterServerCertificateRequest(state);
		//@ assert StateAfterServerCertificateRequest(state) ==> isValidState(state);
		//@ assert ( res == ACCEPT && isValidState(state) ) || (res == REJECT ) ;
		return res;
	case S_HD:
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : Got ServerHelloDone.\n");
#endif
		res = handle_server_hello_done(state, msg, len, direction);
		//@ assert ( res == ACCEPT || res == REJECT );
		//@ assert ( res == ACCEPT ) ==> StateAfterServerHelloDone(state);
		//@ assert StateAfterServerHelloDone(state) ==> isValidState(state);
		//@ assert ( res == ACCEPT && isValidState(state) ) || (res == REJECT ) ;
		return res;
	case C_Cert:
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : Got ClientCertificate.\n");
#endif
		res = handle_client_certificate(state, msg, len, direction);
		//@ assert ( res == ACCEPT || res == REJECT );
		//@ assert ( res == ACCEPT ) ==> StateAfterClientCertificate(state);
		//@ assert StateAfterClientCertificate(state) ==> isValidState(state);
		return res;
	case C_KE:
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : Got ClientKeyExchange.\n");
#endif
		res = handle_client_key_exchange(state, msg, len, direction);
		//@ assert ( res == ACCEPT || res == REJECT );
		//@ assert ( res == ACCEPT ) ==> StateAfterClientKeyExchange(state);
		//@ assert StateAfterClientKeyExchange(state) ==> isValidState(state);
		return res;
	case C_CV:
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : Got ClientCertificateVerify.\n");
#endif
		res = handle_client_certificate_verify(state, msg, len, direction);
		//@ assert ( res == ACCEPT || res == REJECT );
		//@ assert ( res == ACCEPT ) ==> StateAfterClientCertificateVerify(state);
		//@ assert StateAfterClientCertificateVerify(state) ==> isValidState(state);
		return res;
	case ServerNewSessionTicket:
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : Got ServerNewSessionTicket.\n");
#endif
		res = handle_server_new_session_ticket(state, msg, len, direction);
		//@ assert ( res == ACCEPT || res == REJECT );
		//@ assert ( res == ACCEPT ) ==> StateAfterServerNewSessionTicket(state);
		//@ assert StateAfterServerNewSessionTicket(state) ==> isValidState(state);
		return res;
	case S_CCS:
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : Got ServerCCS.\n");
#endif
		res = handle_server_change_cipher_specs(state, msg, len, direction);
		//@ assert ( res == ACCEPT || res == REJECT );
		//@ assert ( res == ACCEPT ) ==> StateAfterServerCCS(state);
		//@ assert StateAfterServerCCS(state) ==> isValidState(state);
		return res;
	case S_Fin:
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : Got ServerFinished.\n");
#endif
		res = handle_server_fin(state, msg, len, direction);
		//@ assert ( res == ACCEPT || res == REJECT );
		//@ assert ( res == ACCEPT ) ==> StateAfterServerFin(state);
		//@ assert StateAfterServerFin(state) ==> isValidState(state);
		return res;
	case C_CCS:
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : Got ClientCCS.\n");
#endif
		res = handle_client_change_cipher_specs(state, msg, len, direction);
		//@ assert ( res == ACCEPT || res == REJECT );
		//@ assert ( res == ACCEPT ) ==> ( StateAfterClientCCS(state) || StateAfterClientCCSLastMsg(state) );
		//@ assert ( StateAfterClientCCS(state) || StateAfterClientCCSLastMsg(state) ) ==> isValidState(state);
		return res;
	case C_Fin:
#ifdef DEBUG
		fprintf(stderr, "STATE MONITOR : Got ClientFinished.\n");
#endif
		res = handle_client_fin(state, msg, len, direction);
		//@ assert ( res == ACCEPT || res == REJECT );
		//@ assert ( res == ACCEPT ) ==> ( StateAfterClientFin(state) || StateAfterClientFinLastMsg(state));
		//@ assert ( StateAfterClientFin(state) || StateAfterClientFinLastMsg(state)) ==> isValidState(state);
		return res;
	default:
		res = REJECT;
		return res;
	}
#ifdef DEBUG
	fprintf(stderr,
			"STATE MONITOR : Got an unknown data type : 0x%x with content type 0x%x.\n",
			msg[0], content_type);
#endif
	res = REJECT;
	//@ assert res == REJECT;
	return res;
}

int ssl3_handle_next_message(STATE_MON *state, unsigned char* msg,
		unsigned int len, int direction, unsigned char content_type) {
	int fragment_length, new_offset, res;

	switch (content_type) {
	case alert:
		fprintf(stderr, "STATE MONITOR : got alert message.\n");
		res = ssl3_next_message(state, msg, len, direction, content_type);
		break;
	case change_cipher_spec:
		fprintf(stderr, "STATE MONITOR : got CCS message.\n");
		res = ssl3_next_message(state, msg, len, direction, content_type);
		break;
	case handshake:
		new_offset = 0;
		res = ACCEPT;
		do {
			fragment_length = msg[new_offset + 1] * 65536
					+ msg[new_offset + 2] * 256 + msg[new_offset + 3];

			if (msg[new_offset + 0] == hello_request) {
				if (msg[new_offset + 1] == 0 && msg[new_offset + 2] == 0
						&& msg[new_offset + 3] == 0) {
					fprintf(stderr,
							"STATE MONITOR : skipped a hello request message.\n");
					res &= ACCEPT;
				} else {
					fprintf(stderr,
							"STATE MONITOR : ERROR : non zero length for a hello request message : %d\n",
							fragment_length);
					res &= REJECT;
				}
			} else {
				res &= ssl3_next_message(state, msg + new_offset,
						fragment_length+4, direction, content_type);
			}
			new_offset += fragment_length + 4;
		} while (new_offset < len);
	        
		break;
	default:
		fprintf(stderr, "STATE MONITOR : FAILURE, unexpected data message during handshake\n");
		res = REJECT;
		break;
	}
	return res;
}
